from typing import Any, Dict, List, Optional, Tuple

ENTRY = Dict[str, Any]


def parse_sib_format(text) -> Tuple[Optional[ENTRY], List[ENTRY]]:
    """
    Parse SIB/Swiss-Prot format data into a structured dictionary.

    Args:
        text (str): The text in SIB/Swiss-Prot format

    Returns:
        dict: A dictionary with entry IDs as keys and parsed data as values
    """
    # Split the text into entries (separated by //)
    entries = text.split("//\n")
    header = None

    # Initialize results dictionary
    results = []

    # Parse each entry
    for entry in entries:
        if not entry.strip():
            continue

        # Initialize dictionary for current entry
        current_entry = {}
        current_code = None

        # Process each line
        for line in entry.strip().split("\n"):
            if not line.strip():
                continue

            # Check if this is a new field (starts with a 2-letter code followed by space)
            if len(line) > 2 and line[2] == " ":
                current_code = line[0:2]
                # Remove the code and the following space(s)
                value = line[3:].strip()

                # Initialize as list if needed for multi-line fields
                if current_code not in current_entry:
                    current_entry[current_code] = []

                current_entry[current_code].append(value)

            # Continuation of previous field
            elif current_code is not None:
                # Handle continuation lines (typically indented)
                if current_code == "CC":
                    # For comments, preserve the indentation
                    current_entry[current_code].append(line)
                else:
                    # For other fields, strip and append
                    current_entry[current_code].append(line.strip())

        # Combine multiline comments; e.g
        # -!- ...
        #     ...
        # -!- ...
        ccs = current_entry.get("CC", [])
        new_ccs = []
        for cc in ccs:
            if not cc.startswith("-!-") and new_ccs:
                new_ccs[-1] += " " + cc
            else:
                new_ccs.append(cc)
        current_entry["CC"] = new_ccs
        for k, vs in current_entry.items():
            if k != "CC":
                combined = "".join(vs)
                combined = combined.strip()
                if combined.endswith("."):
                    combined = combined.split(".")
                    combined = [c.strip() for c in combined if c.strip()]
                    if k == "DE":
                        combined = combined[0]
                current_entry[k] = combined

        if "ID" in current_entry:
            results.append(current_entry)
        else:
            header = current_entry

    return header, results


# Example usage:
# data = parse_sib_format(text)
# for entry_id, entry_data in data.items():
#     print(f"Entry: {entry_id}")
#     for code, values in entry_data.items():
#         print(f"  {code}: {values}")



import logging
from typing import Optional

import requests

from linkml_store.api import Collection, Database
from linkml_store.api.config import CollectionConfig
from linkml_store.api.queries import Query, QueryResult
from linkml_store.api.stores.solr.solr_collection import SolrCollection

logger = logging.getLogger(__name__)


class SolrDatabase(Database):
    base_url: str
    collection_class = SolrCollection
    use_cores: bool = False

    def __init__(self, handle: Optional[str] = None, **kwargs):
        if handle.startswith("solr:"):
            self.base_url = handle.replace("solr:", "")
        else:
            self.base_url = handle
        super().__init__(handle=handle, **kwargs)

    def get_collection(self, name: str, create_if_not_exists=True, **kwargs) -> "Collection":
        if not self._collections:
            self.init_collections()

        if name not in self._collections.keys():
            if create_if_not_exists:
                self._collections[name] = self.create_collection(name)
            else:
                raise KeyError(f"Collection {name} does not exist")

        return self._collections[name]

    def create_collection(
        self, name: str, alias: Optional[str] = None, metadata: Optional[CollectionConfig] = None, **kwargs
    ) -> Collection:
        if not name:
            raise ValueError(f"Collection name must be provided: alias: {alias} metadata: {metadata}")

        collection_cls = self.collection_class
        collection = collection_cls(name=name, alias=alias, parent=self, metadata=metadata)

        if not self._collections:
            self._collections = {}

        if not alias:
            alias = name

        self._collections[alias] = collection
        return collection

    def init_collections(self):
        if self._collections is None:
            self._collections = {}
        if self.metadata.collection_type_slot:
            response = requests.get(
                f"{self.base_url}/select",
                params={
                    "q": "*:*",
                    "wt": "json",
                    "rows": 0,
                    "facet": "true",
                    "facet.field": self.metadata.collection_type_slot,
                    "facet.limit": -1,
                },
            )
            response.raise_for_status()
            data = response.json()
            coll_names = data["facet_counts"]["facet_fields"][self.metadata.collection_type_slot]
            coll_names = coll_names[::2]
            for coll_name in coll_names:
                self.create_collection(coll_name)
        else:
            self.create_collection("default")

    def query(self, query: Query, **kwargs) -> QueryResult:
        collection_name = query.from_table
        collection = self.get_collection(collection_name)
        return collection.query(query, **kwargs)



# solr_collection.py

import logging
from copy import copy
from typing import Any, Dict, List, Optional, Union, Tuple

import requests

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT
from linkml_store.api.queries import Query, QueryResult

logger = logging.getLogger(__name__)


class SolrCollection(Collection):

    @property
    def _collection_base(self) -> str:
        if self.parent.use_cores:
            base_url = f"{self.parent.base_url}/{self.alias}"
        else:
            base_url = self.parent.base_url
        return base_url

    def search(
        self,
        query: str,
        where: Optional[Any] = None,
        index_name: Optional[str] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> QueryResult:
        if index_name is None:
            index_name = "edismax"
        qfs = self.parent.metadata.searchable_slots
        #if not qfs:
        #    raise ValueError("No searchable slots configured for Solr collection")
        if not qfs:
            qfs = list(self.class_definition().attributes.keys())
        solr_query = self._build_solr_query(where, search_term=query, extra={"defType": index_name, "qf": qfs})
        logger.info(f"Querying Solr collection {self.alias} with query: {solr_query}")

        response = requests.get(f"{self._collection_base}/select", params=solr_query)
        response.raise_for_status()

        data = response.json()
        num_rows = data["response"]["numFound"]
        rows = data["response"]["docs"]
        ranked_rows = [(1.0, row) for row in rows]
        return QueryResult(query=where, search_term=query, num_rows=num_rows, rows=rows, ranked_rows=ranked_rows)

    def query(self, query: Query, **kwargs) -> QueryResult:
        solr_query = self._build_solr_query(query)
        logger.info(f"Querying Solr collection {self.alias} with query: {solr_query}")

        response = requests.get(f"{self._collection_base}/select", params=solr_query)
        response.raise_for_status()

        data = response.json()
        logger.debug(f"Response: {data}")
        num_rows = data["response"]["numFound"]
        rows = data["response"]["docs"]

        return QueryResult(query=query, num_rows=num_rows, rows=rows)

    def query_facets(
        self,
        where: Optional[Dict] = None,
        facet_columns: List[Union[str, Tuple[str, ...]]] = None,
        facet_limit=DEFAULT_FACET_LIMIT,
        facet_min_count: int = 1,
        **kwargs,
    ) -> Dict[Union[str, Tuple[str, ...]], List[Tuple[Any, int]]]:
        """
        Query facet counts for fields or field combinations.
        
        :param where: Filter conditions
        :param facet_columns: List of fields to facet on. Elements can be:
                            - Simple strings for single field facets
                            - Tuples of strings for field combinations (pivot facets)
        :param facet_limit: Maximum number of facet values to return
        :param facet_min_count: Minimum count for facet values to be included
        :return: Dictionary mapping fields or field tuples to lists of (value, count) tuples
        """
        solr_query = self._build_solr_query(where)
        
        # Separate single fields and tuple fields
        single_fields = []
        tuple_fields = []
        
        if facet_columns:
            for field in facet_columns:
                if isinstance(field, str):
                    single_fields.append(field)
                elif isinstance(field, tuple):
                    tuple_fields.append(field)
        
        # Process regular facets
        results = {}
        if single_fields:
            solr_query["facet"] = "true"
            solr_query["facet.field"] = single_fields
            solr_query["facet.limit"] = facet_limit
            solr_query["facet.mincount"] = facet_min_count
            
            logger.info(f"Querying Solr collection {self.alias} for facets with query: {solr_query}")
            response = requests.get(f"{self._collection_base}/select", params=solr_query)
            response.raise_for_status()
            
            data = response.json()
            facet_counts = data["facet_counts"]["facet_fields"]
            
            for facet_field, counts in facet_counts.items():
                results[facet_field] = list(zip(counts[::2], counts[1::2]))
        
        # Process pivot facets for tuple fields
        if tuple_fields:
            # TODO: Add a warning if Solr < 4.0, when this was introduced
            for field_tuple in tuple_fields:
                # Create a query for this specific field tuple
                pivot_query = self._build_solr_query(where)
                pivot_query["facet"] = "true"
                
                # Create pivot facet
                field_str = ','.join(field_tuple)
                pivot_query["facet.pivot"] = field_str
                pivot_query["facet.pivot.mincount"] = facet_min_count
                pivot_query["facet.limit"] = facet_limit
                
                logger.info(f"Querying Solr collection {self.alias} for pivot facets with query: {pivot_query}")
                response = requests.get(f"{self._collection_base}/select", params=pivot_query)
                response.raise_for_status()
                
                data = response.json()
                pivot_facets = data.get("facet_counts", {}).get("facet_pivot", {})
                
                # Process pivot facets into the same format as MongoDB results
                field_str = ','.join(field_tuple)
                pivot_data = pivot_facets.get(field_str, [])
                
                # Build a list of tuples (field values, count)
                pivot_results = []
                self._process_pivot_facets(pivot_data, [], pivot_results, field_tuple)
                
                results[field_tuple] = pivot_results
        
        return results
        
    def _process_pivot_facets(self, pivot_data, current_values, results, field_tuple):
        """
        Recursively process pivot facet results to extract combinations of field values.
        
        :param pivot_data: The pivot facet data from Solr
        :param current_values: The current path of values in the recursion
        :param results: The result list to populate
        :param field_tuple: The original field tuple for reference
        """
        for item in pivot_data:
            # Add the current field value
            value = item.get("value")
            count = item.get("count", 0)
            
            # Update the current path with this value
            values = current_values + [value]
            
            # If we have all the fields from the tuple, add a result
            if len(values) == len(field_tuple):
                # Create a tuple of values corresponding to the field tuple
                results.append((tuple(values), count))
            
            # Process child pivot fields recursively
            pivot = item.get("pivot", [])
            if pivot and len(values) < len(field_tuple):
                self._process_pivot_facets(pivot, values, results, field_tuple)

    def _build_solr_query(
        self, query: Union[Query, Dict], search_term="*:*", extra: Optional[Dict] = None
    ) -> Dict[str, Any]:
        solr_query = {}
        if query is None:
            query = {}

        if isinstance(query, Query):
            where = query.where_clause
            solr_query["fq"] = self._build_solr_where_clause(where)

            if query.select_cols:
                solr_query["fl"] = ",".join(query.select_cols)

            if query.limit:
                solr_query["rows"] = query.limit

            if query.offset:
                solr_query["start"] = query.offset

        elif isinstance(query, dict):
            solr_query["fq"] = self._build_solr_where_clause(query)

        solr_query["wt"] = "json"
        if "q" not in solr_query:
            solr_query["q"] = search_term
        if extra:
            solr_query.update(extra)
        logger.info(f"Built Solr query: {solr_query}")
        return solr_query

    def _build_solr_where_clause(self, where_clause: Dict) -> str:
        if where_clause is None:
            where_clause = {}
        conditions = []
        if self.parent.metadata.collection_type_slot:
            where_clause = copy(where_clause)
            where_clause[self.parent.metadata.collection_type_slot] = self.alias
        for field, value in where_clause.items():
            if not isinstance(value, (list, tuple)):
                value = [value]
            value = [f'"{v}"' if isinstance(v, str) else str(v) for v in value]
            if len(value) > 1:
                conditions.append(f"{field}:({' '.join(value)})")
            else:
                conditions.append(f"{field}:{value[0]}")

        return " AND ".join(conditions)



"""
Heatmap visualization module for LinkML data.

This module provides functions to generate heatmaps from pandas DataFrames or tabular data files.
"""

import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.colors import LinearSegmentedColormap
from scipy.cluster import hierarchy
from scipy.spatial import distance

from linkml_store.utils.format_utils import Format, load_objects, write_output

logger = logging.getLogger(__name__)


def create_heatmap(
    data: pd.DataFrame,
    x_column: str,
    y_column: str,
    value_column: Optional[str] = None,
    minimum_value: Optional[float] = None,
    title: Optional[str] = None,
    figsize: Tuple[int, int] = (10, 8),
    cmap: Union[str, LinearSegmentedColormap] = "YlGnBu",
    annot: bool = True,
    fmt: Optional[str] = None,  # Dynamically determined based on data
    linewidths: float = 0.5,
    linecolor: str = "white",
    square: bool = False,
    output_file: Optional[str] = None,
    dpi: int = 300,
    missing_value: Any = np.nan,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    robust: bool = False,
    remove_duplicates: bool = False,
    font_size: int = 10,
    cluster: Union[bool, Literal["both", "x", "y"]] = False,
    cluster_method: str = "complete",  # linkage method: complete, average, single, etc.
    cluster_metric: str = "euclidean",  # distance metric: euclidean, cosine, etc.
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Create a heatmap from a pandas DataFrame.

    Args:
        data: Input DataFrame containing the data to plot
        x_column: Column to use for x-axis categories
        y_column: Column to use for y-axis categories
        value_column: Column containing values for the heatmap. If None, frequency counts will be used.
        minimum_value: Minimum value to include in the heatmap
        title: Title for the heatmap
        figsize: Figure size as (width, height) in inches
        cmap: Colormap for the heatmap
        annot: Whether to annotate cells with values
        fmt: String formatting code for annotations (auto-detected if None)
        linewidths: Width of lines between cells
        linecolor: Color of lines between cells
        square: Whether to make cells square
        output_file: File path to save the figure (optional)
        dpi: Resolution for saved figure
        missing_value: Value to use for missing data (defaults to NaN)
        vmin: Minimum value for colormap scaling
        vmax: Maximum value for colormap scaling
        robust: If True, compute colormap limits using robust quantiles instead of min/max
        remove_duplicates: If True, removes duplicate rows before creating the heatmap
        font_size: Font size for annotations
        cluster: Whether and which axes to cluster:
                - False: No clustering (default)
                - True or "both": Cluster both x and y axes
                - "x": Cluster only x-axis
                - "y": Cluster only y-axis
        cluster_method: Linkage method for hierarchical clustering
                      (e.g., "single", "complete", "average", "ward")
        cluster_metric: Distance metric for clustering (e.g., "euclidean", "correlation", "cosine")
        **kwargs: Additional keyword arguments to pass to seaborn's heatmap function

    Returns:
        Tuple containing the figure and axes objects
    """
    # Validate input
    if x_column not in data.columns:
        raise ValueError(f"x_column '{x_column}' not found in DataFrame columns: {list(data.columns)}")
    if y_column not in data.columns:
        raise ValueError(f"y_column '{y_column}' not found in DataFrame columns: {list(data.columns)}")
    if value_column and value_column not in data.columns:
        raise ValueError(f"value_column '{value_column}' not found in DataFrame columns: {list(data.columns)}")
    

    cols = [x_column, y_column]
    if value_column:
        cols.append(value_column)

    # select cols from data
    data = data[cols]  

    if any(isinstance(val, (list, set, tuple)) for val in data[x_column].dropna().head(100)):
        logger.info(f"Exploding list values in x_column '{x_column}'")
        data = data.explode(x_column).dropna(subset=[x_column])

    if any(isinstance(val, (list, set, tuple)) for val in data[y_column].dropna().head(100)):
        logger.info(f"Exploding list values in y_column '{y_column}'")
        data = data.explode(y_column).dropna(subset=[y_column])

    if value_column:
        if any(isinstance(val, (list, set, tuple)) for val in data[value_column].dropna().head(100)):
            logger.info(f"Exploding list values in value_column '{value_column}'")
            data = data.explode(value_column).dropna(subset=[value_column])

    # Remove duplicates by default (assume they're accidents unless user overrides)
    if remove_duplicates:
        data = data.drop_duplicates()

    if value_column and minimum_value is not None:
        data = data[data[value_column] >= minimum_value]
    
    # Prepare the data
    if value_column:
        # Use the provided value column
        # print(data.head())
        pivot_data = data.pivot_table(
            index=y_column, 
            columns=x_column, 
            values=value_column, 
            aggfunc='mean',
            fill_value=missing_value
        )
    else:
        # Use frequency counts
        cross_tab = pd.crosstab(data[y_column], data[x_column])
        pivot_data = cross_tab
    
    # Auto-detect format string if not provided
    if fmt is None:
        # Check if the pivot table contains integers only
        if pivot_data.dtypes.apply(lambda x: pd.api.types.is_integer_dtype(x)).all():
            fmt = 'd'  # Integer format
        else:
            fmt = '.1f'  # One decimal place for floats
    
    # Make sure all cells have a reasonable minimum size
    min_height = max(4, 80 / len(pivot_data.index) if len(pivot_data.index) > 0 else 10)
    min_width = max(4, 80 / len(pivot_data.columns) if len(pivot_data.columns) > 0 else 10)
    
    # Adjust figure size based on the number of rows and columns 
    adjusted_height = max(figsize[1], min_height * len(pivot_data.index) / 10)
    adjusted_width = max(figsize[0], min_width * len(pivot_data.columns) / 10)
    adjusted_figsize = (adjusted_width, adjusted_height)
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=adjusted_figsize)
    
    # Apply clustering if requested
    row_linkage = None
    col_linkage = None
    
    if cluster:
        cluster_axes = cluster
        if cluster_axes is True:
            cluster_axes = "both"
            
        # Fill NAs for clustering
        pivot_data_for_clustering = pivot_data.fillna(0)
        
        # Cluster rows (y-axis)
        if cluster_axes in ["both", "y"]:
            try:
                # Calculate distance matrix and linkage for rows
                row_distances = distance.pdist(pivot_data_for_clustering.values, metric=cluster_metric)
                row_linkage = hierarchy.linkage(row_distances, method=cluster_method)
                
                # Reorder rows based on clustering
                row_dendrogram = hierarchy.dendrogram(row_linkage, no_plot=True)
                row_order = row_dendrogram['leaves']
                pivot_data = pivot_data.iloc[row_order]
                
                logger.info(f"Applied clustering to rows using {cluster_method} linkage and {cluster_metric} metric")
            except Exception as e:
                logger.warning(f"Failed to cluster rows: {e}")
        
        # Cluster columns (x-axis)
        if cluster_axes in ["both", "x"]:
            try:
                # Calculate distance matrix and linkage for columns
                col_distances = distance.pdist(pivot_data_for_clustering.values.T, metric=cluster_metric)
                col_linkage = hierarchy.linkage(col_distances, method=cluster_method)
                
                # Reorder columns based on clustering
                col_dendrogram = hierarchy.dendrogram(col_linkage, no_plot=True)
                col_order = col_dendrogram['leaves']
                pivot_data = pivot_data.iloc[:, col_order]
                
                logger.info(f"Applied clustering to columns using {cluster_method} linkage and {cluster_metric} metric")
            except Exception as e:
                logger.warning(f"Failed to cluster columns: {e}")

    # Create the heatmap
    sns.heatmap(
        pivot_data,
        cmap=cmap,
        annot=annot,
        fmt=fmt,
        linewidths=linewidths,
        linecolor=linecolor,
        square=square,
        vmin=vmin,
        vmax=vmax,
        robust=robust,
        ax=ax,
        annot_kws={'fontsize': font_size},
        **kwargs
    )
    
    # Set title if provided
    if title:
        ax.set_title(title, fontsize=font_size + 4)
    
    # Improve display of tick labels
    plt.xticks(rotation=45, ha="right", fontsize=font_size)
    plt.yticks(rotation=0, fontsize=font_size)
    
    # Add grid lines to make the table more readable
    ax.grid(False)
    
    # Improve contrast for better readability
    for _, spine in ax.spines.items():
        spine.set_visible(True)
        spine.set_color('black')
        spine.set_linewidth(1)
    
    # Adjust layout
    plt.tight_layout()
    
    # Save the figure if output file is specified
    if output_file:
        output_path = Path(output_file)
        output_dir = output_path.parent
        if not output_dir.exists():
            output_dir.mkdir(parents=True, exist_ok=True)
        plt.savefig(output_file, dpi=dpi, bbox_inches="tight")
        logger.info(f"Heatmap saved to {output_file}")
    
    return fig, ax


def heatmap_from_file(
    file_path: str,
    x_column: str,
    y_column: str,
    value_column: Optional[str] = None,
    format: Optional[Union[Format, str]] = None,
    compression: Optional[str] = None,
    output_file: Optional[str] = None,
    remove_duplicates: bool = True,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Create a heatmap from a file (CSV, TSV, etc.).

    Args:
        file_path: Path to the input file or "-" for stdin
        x_column: Column to use for x-axis categories
        y_column: Column to use for y-axis categories
        value_column: Column containing values for the heatmap. If None, frequency counts will be used.
        format: Format of the input file (auto-detected if None)
        compression: Compression format ('gz' or 'tgz')
        output_file: File path to save the figure (optional)
        remove_duplicates: If True, removes duplicate rows before creating the heatmap
        **kwargs: Additional arguments to pass to create_heatmap

    Returns:
        Tuple containing the figure and axes objects
    """
    # Handle stdin input safely
    import sys
    import io
    import pandas as pd
    import click
    
    # Load the data
    if file_path == "-":
        # Read directly from stdin since format_utils will use sys.stdin which may already be consumed
        if not format or str(format).lower() in ['csv', 'tsv']:
            # Default to CSV if no format specified
            delimiter = ',' if not format or str(format).lower() == 'csv' else '\t'
            df = pd.read_csv(sys.stdin, delimiter=delimiter)
        else:
            # Try to use format_utils but with a backup plan
            try:
                objs = load_objects(file_path, format=format, compression=compression)
                df = pd.DataFrame(objs)
            except ValueError as e:
                if "I/O operation on closed file" in str(e):
                    logger.warning("Could not read from stdin. It may have been consumed already.")
                    raise click.UsageError("Error reading from stdin. Please provide a file path or ensure stdin has data.")
                else:
                    raise
    else:
        # For regular files, use format_utils as normal
        if (not format or format in ["csv", "tsv"]) and not compression:
            df = pd.read_csv(file_path)
        else:
            objs = load_objects(file_path, format=format, compression=compression)
            df = pd.DataFrame(objs)

    # Create the heatmap
    return create_heatmap(
        data=df,
        x_column=x_column,
        y_column=y_column,
        value_column=value_column,
        output_file=output_file,
        remove_duplicates=remove_duplicates,
        **kwargs
    )


def export_heatmap_data(
    data: pd.DataFrame,
    x_column: str,
    y_column: str,
    value_column: Optional[str] = None,
    output_file: Optional[str] = None,
    format: Union[Format, str] = Format.CSV,
    missing_value: Any = np.nan,
    remove_duplicates: bool = True,
) -> pd.DataFrame:
    """
    Export heatmap data to a file or return it as a DataFrame.

    Args:
        data: Input DataFrame containing the data
        x_column: Column to use for x-axis categories
        y_column: Column to use for y-axis categories
        value_column: Column containing values for the heatmap. If None, frequency counts will be used.
        output_file: File path to save the data (optional)
        format: Output format for the file
        missing_value: Value to use for missing data
        remove_duplicates: If True, removes duplicate rows before creating the pivot table

    Returns:
        DataFrame containing the pivot table data
    """
    # Remove duplicates by default (assume they're accidents unless user overrides)
    if remove_duplicates:
        # Keep the first occurrence of each x_column, y_column combination
        data = data.drop_duplicates(subset=[x_column, y_column])

    # Prepare the data
    if value_column:
        # Use the provided value column
        pivot_data = data.pivot_table(
            index=y_column, 
            columns=x_column, 
            values=value_column, 
            aggfunc='mean',
            fill_value=missing_value
        )
    else:
        # Use frequency counts
        cross_tab = pd.crosstab(data[y_column], data[x_column])
        pivot_data = cross_tab
    
    # Reset index to make the y_column a regular column
    result_df = pivot_data.reset_index()
    
    # Write to file if output_file is provided
    if output_file:
        # Convert to records format for writing
        records = result_df.to_dict(orient='records')
        write_output(records, format=format, target=output_file)
        logger.info(f"Heatmap data saved to {output_file}")
    
    return result_df


"""
inference engine package.
"""

from linkml_store.inference.inference_config import InferenceConfig
from linkml_store.inference.inference_engine import InferenceEngine
from linkml_store.inference.inference_engine_registry import get_inference_engine

__all__ = [
    "InferenceEngine",
    "InferenceConfig",
    "get_inference_engine",
]



import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, List, Optional, TextIO, Union

import yaml
from llm import get_key
from pydantic import BaseModel

from linkml_store.api.collection import OBJECT, Collection
from linkml_store.inference.inference_config import Inference, InferenceConfig, LLMConfig
from linkml_store.inference.inference_engine import InferenceEngine, ModelSerialization
from linkml_store.utils.object_utils import select_nested

logger = logging.getLogger(__name__)

MAX_ITERATIONS = 5
DEFAULT_NUM_EXAMPLES = 20
DEFAULT_MMR_RELEVANCE_FACTOR = 0.8

SYSTEM_PROMPT = """
You are a {llm_config.role}, your task is to infer the YAML
object output given the YAML object input. I will provide you
with a collection of examples that will provide guidance both
on the desired structure of the response, as well as the kind
of content.

You should return ONLY valid YAML in your response.
"""


class TrainedModel(BaseModel, extra="forbid"):
    rag_collection_rows: List[OBJECT]
    index_rows: List[OBJECT]
    config: Optional[InferenceConfig] = None


class RAGInference(Inference):
    iterations: int = 0


@dataclass
class RAGInferenceEngine(InferenceEngine):
    """
    AI Retrieval Augmented Generation (RAG) based predictor.


    >>> from linkml_store.api.client import Client
    >>> from linkml_store.utils.format_utils import Format
    >>> from linkml_store.inference.inference_config import LLMConfig
    >>> client = Client()
    >>> db = client.attach_database("duckdb", alias="test")
    >>> db.import_database("tests/input/countries/countries.jsonl", Format.JSONL, collection_name="countries")
    >>> db.list_collection_names()
    ['countries']
    >>> collection = db.get_collection("countries")
    >>> features = ["name"]
    >>> targets = ["code", "capital", "continent", "languages"]
    >>> llm_config = LLMConfig(model_name="gpt-4o-mini",)
    >>> config = InferenceConfig(target_attributes=targets, feature_attributes=features, llm_config=llm_config)
    >>> ie = RAGInferenceEngine(config=config)
    >>> ie.load_and_split_data(collection)
    >>> ie.initialize_model()
    >>> prediction = ie.derive({"name": "Uruguay"})
    >>> prediction.predicted_object
    {'capital': 'Montevideo', 'code': 'UY', 'continent': 'South America', 'languages': ['Spanish']}

    The "model" can be saved for later use:

    >>> ie.export_model("tests/output/countries.rag_model.json")

    Note in this case the model is not the underlying LLM, but the "RAG Model" which is the vectorized
    representation of training set objects.

    """

    _model: "llm.Model" = None  # noqa: F821

    rag_collection: Collection = None

    PERSIST_COLS: ClassVar[List[str]] = [
        "config",
    ]

    def __post_init__(self):
        if not self.config:
            self.config = InferenceConfig()
        if not self.config.llm_config:
            self.config.llm_config = LLMConfig()

    @property
    def model(self) -> "llm.Model":  # noqa: F821
        import llm

        if self._model is None:
            self._model = llm.get_model(self.config.llm_config.model_name)
            if self._model.needs_key:
                key = get_key(None, key_alias=self._model.needs_key)
                self._model.key = key

        return self._model

    def initialize_model(self, **kwargs):
        logger.info(f"Initializing model {self.model}")
        if self.training_data:
            rag_collection = self.training_data.collection
            rag_collection.attach_indexer("llm", auto_index=False)
            self.rag_collection = rag_collection

    def object_to_text(self, object: OBJECT) -> str:
        return yaml.dump(object)

    def derive(
        self, object: OBJECT, iteration=0, additional_prompt_texts: Optional[List[str]] = None
    ) -> Optional[RAGInference]:
        import llm
        from tiktoken import encoding_for_model

        from linkml_store.utils.llm_utils import get_token_limit, render_formatted_text

        model: llm.Model = self.model
        model_name = self.config.llm_config.model_name
        feature_attributes = self.config.feature_attributes
        target_attributes = self.config.target_attributes
        num_examples = self.config.llm_config.number_of_few_shot_examples or DEFAULT_NUM_EXAMPLES
        query_text = self.object_to_text(object)
        mmr_relevance_factor = DEFAULT_MMR_RELEVANCE_FACTOR
        if not self.rag_collection:
            # TODO: zero-shot mode
            examples = []
        else:
            if not self.rag_collection.indexers:
                raise ValueError("RAG collection must have an indexer attached")
            logger.info(f"Searching {self.rag_collection.alias} for examples for: {query_text}")
            rs = self.rag_collection.search(
                query_text, limit=num_examples, index_name="llm", mmr_relevance_factor=mmr_relevance_factor
            )
            examples = rs.rows
            logger.info(f"Found {len(examples)} examples")
            if not examples:
                raise ValueError(f"No examples found for {query_text}; size = {self.rag_collection.size()}")
        prompt_clauses = []
        this_feature_attributes = feature_attributes
        if not this_feature_attributes:
            this_feature_attributes = list(set(object.keys()) - set(target_attributes))
        query_obj = select_nested(object, this_feature_attributes)
        query_text = self.object_to_text(query_obj)
        for example in examples:
            this_feature_attributes = feature_attributes
            if not this_feature_attributes:
                this_feature_attributes = list(set(example.keys()) - set(target_attributes))
            if not this_feature_attributes:
                raise ValueError(f"No feature attributes found in example {example}")
            input_obj = select_nested(example, this_feature_attributes)
            input_obj_text = self.object_to_text(input_obj)
            if input_obj_text == query_text:
                continue
                # raise ValueError(
                #    f"Query object {query_text} is the same as example object {input_obj_text}\n"
                #    "This indicates possible test data leakage\n."
                #    "TODO: allow an option that allows user to treat this as a basic lookup\n"
                # )
            output_obj = select_nested(example, target_attributes)
            prompt_clause = (
                "---\nExample:\n" f"## INPUT:\n{input_obj_text}\n" f"## OUTPUT:\n{self.object_to_text(output_obj)}\n"
            )
            prompt_clauses.append(prompt_clause)

        system_prompt = SYSTEM_PROMPT.format(llm_config=self.config.llm_config)
        system_prompt += "\n".join(additional_prompt_texts or [])
        prompt_end = "---\nQuery:\n" f"## INPUT:\n{query_text}\n" "## OUTPUT:\n"

        def make_text(texts: List[str]):
            return "\n".join(texts) + prompt_end

        try:
            encoding = encoding_for_model(model_name)
        except KeyError:
            encoding = encoding_for_model("gpt-4")
        token_limit = get_token_limit(model_name)
        prompt = render_formatted_text(
            make_text, values=prompt_clauses, encoding=encoding, token_limit=token_limit, additional_text=system_prompt
        )
        logger.info(f"Prompt: {prompt}")
        response = model.prompt(prompt, system=system_prompt)
        yaml_str = response.text()
        logger.info(f"Response: {yaml_str}")
        predicted_object = self._parse_yaml_payload(yaml_str, strict=True)
        if self.config.validate_results:
            base_collection = self.training_data.base_collection
            errs = list(base_collection.iter_validate_collection([predicted_object]))
            if errs:
                print(f"{iteration} // FAILED TO VALIDATE: {yaml_str}")
                print(f"PARSED: {predicted_object}")
                print(f"ERRORS: {errs}")
                if iteration > MAX_ITERATIONS:
                    raise ValueError(f"Validation errors: {errs}")
                extra_texts = [
                    "Make sure results conform to the schema. Previously you provided:\n",
                    yaml_str,
                    "\nThis was invalid.\n",
                    "Validation errors:\n",
                ] + [self.object_to_text(e) for e in errs]
                return self.derive(object, iteration=iteration + 1, additional_prompt_texts=extra_texts)
        return RAGInference(predicted_object=predicted_object, iterations=iteration + 1, query=object)

    def _parse_yaml_payload(self, yaml_str: str, strict=False) -> Optional[OBJECT]:
        if "```" in yaml_str:
            yaml_str = yaml_str.split("```")[1].strip()
            if yaml_str.startswith("yaml"):
                yaml_str = yaml_str[4:].strip()
        try:
            return yaml.safe_load(yaml_str)
        except Exception as e:
            if strict:
                raise e
            logger.error(f"Error parsing YAML: {yaml_str}\n{e}")
            return None

    def export_model(
        self, output: Optional[Union[str, Path, TextIO]], model_serialization: ModelSerialization = None, **kwargs
    ):
        self.save_model(output)

    def save_model(self, output: Union[str, Path]) -> None:
        """
        Save the trained model and related data to a file.

        :param output: Path to save the model
        """

        # trigger index
        _qr = self.rag_collection.search("*", limit=1)
        assert len(_qr.ranked_rows) > 0

        rows = self.rag_collection.find(limit=-1).rows

        indexers = self.rag_collection.indexers
        assert len(indexers) == 1
        ix = self.rag_collection.indexers["llm"]
        ix_coll = self.rag_collection.parent.get_collection(self.rag_collection.get_index_collection_name(ix))

        ix_rows = ix_coll.find(limit=-1).rows
        assert len(ix_rows) > 0
        tm = TrainedModel(rag_collection_rows=rows, index_rows=ix_rows, config=self.config)
        # tm = TrainedModel(rag_collection_rows=rows, index_rows=ix_rows)
        with open(output, "w", encoding="utf-8") as f:
            json.dump(tm.model_dump(), f)

    @classmethod
    def load_model(cls, file_path: Union[str, Path]) -> "RAGInferenceEngine":
        """
        Load a trained model and related data from a file.

        :param file_path: Path to the saved model
        :return: SklearnInferenceEngine instance with loaded model
        """
        with open(file_path, "r", encoding="utf-8") as f:
            model_data = json.load(f)
        tm = TrainedModel(**model_data)
        from linkml_store.api import Client

        client = Client()
        db = client.attach_database("duckdb", alias="training")
        db.store({"data": tm.rag_collection_rows})
        collection = db.get_collection("data")
        ix = collection.attach_indexer("llm", auto_index=False)
        assert ix.name
        ix_coll_name = collection.get_index_collection_name(ix)
        assert ix_coll_name
        ix_coll = db.get_collection(ix_coll_name, create_if_not_exists=True)
        ix_coll.insert(tm.index_rows)
        ie = cls(config=tm.config)
        ie.rag_collection = collection
        return ie



import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query, QueryResult
from linkml_store.api.types import DatabaseType
from linkml_store.utils.query_utils import mongo_query_to_match_function

logger = logging.getLogger(__name__)


class FileSystemCollection(Collection[DatabaseType]):
    path: Optional[Path] = None
    file_format: Optional[str] = None
    encoding: Optional[str] = None
    _objects_list: List[OBJECT] = None
    _object_map: Dict[str, OBJECT] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        parent: DatabaseType = self.parent
        if not self.path:
            if self.parent:
                self.path = Path(parent.directory_path)
        self._objects_list = []
        self._object_map = {}
        if not self.file_format:
            self.file_format = "json"

    @property
    def path_to_file(self):
        return Path(self.parent.directory_path) / f"{self.alias}.{self.file_format}"

    @property
    def objects_as_list(self) -> List[OBJECT]:
        if self._object_map:
            return list(self._object_map.values())
        else:
            return self._objects_list

    def _set_objects(self, objs: List[OBJECT]):
        pk = self.identifier_attribute_name
        if pk:
            self._object_map = {obj[pk]: obj for obj in objs}
            self._objects_list = []
        else:
            self._objects_list = objs
            self._object_map = {}

    def commit(self):
        path = self.path_to_file
        if not path:
            raise ValueError("Path not set")
        path.parent.mkdir(parents=True, exist_ok=True)
        self._save(path)

    def _save(self, path: Path):
        encoding = self.encoding or "utf-8"
        fmt = self.file_format or "json"
        mode = "w"
        if fmt == "parquet":
            mode = "wb"
            encoding = None
        with open(path, mode, encoding=encoding) as stream:
            if fmt == "json":
                import json

                json.dump(self.objects_as_list, stream, indent=2)
            elif fmt == "jsonl":
                import jsonlines

                writer = jsonlines.Writer(stream)
                writer.write_all(self.objects_as_list)
            elif fmt == "yaml":
                import yaml

                yaml.dump_all(self.objects_as_list, stream)
            elif fmt == "parquet":
                import pandas as pd
                import pyarrow
                import pyarrow.parquet as pq

                df = pd.DataFrame(self.objects_as_list)
                table = pyarrow.Table.from_pandas(df)
                pq.write_table(table, stream)
            elif fmt in {"csv", "tsv"}:
                import csv

                delimiter = "\t" if fmt == "tsv" else ","
                fieldnames = list(self.objects_as_list[0].keys())
                for obj in self.objects_as_list[1:]:
                    fieldnames.extend([k for k in obj.keys() if k not in fieldnames])
                writer = csv.DictWriter(stream, fieldnames=fieldnames, delimiter=delimiter)
                writer.writeheader()
                for obj in self.objects_as_list:
                    writer.writerow(obj)
            else:
                raise ValueError(f"Unsupported file format: {fmt}")

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        if not isinstance(objs, list):
            objs = [objs]
        if not objs:
            return
        pk = self.identifier_attribute_name
        if pk:
            for obj in objs:
                if pk not in obj:
                    raise ValueError(f"Primary key {pk} not found in object {obj}")
                pk_val = obj[pk]
                self._object_map[pk_val] = obj
        else:
            self._objects_list.extend(objs)

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> Optional[int]:
        if not isinstance(objs, list):
            objs = [objs]
        if not objs:
            return 0
        pk = self.identifier_attribute_name
        n = 0
        if pk:
            for obj in objs:
                pk_val = obj[pk]
                if pk_val in self._object_map:
                    del self._object_map[pk_val]
                    n += 1
        else:
            n = len(objs)
            self._objects_list = [o for o in self._objects_list if o not in objs]
            n = n - len(objs)
        return n

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> Optional[int]:
        logger.info(f"Deleting from {self.target_class_name} where: {where}")
        if where is None:
            where = {}

        def matches(obj: OBJECT):
            for k, v in where.items():
                if obj.get(k) != v:
                    return False
            return True

        print(type(self))
        print(self)
        print(vars(self))
        curr_objects = [o for o in self.objects_as_list if not matches(o)]
        self._set_objects(curr_objects)

    def query(self, query: Query, limit: Optional[int] = None, offset: Optional[int] = None, **kwargs) -> QueryResult:
        limit = limit or query.limit
        offset = offset or query.offset
        if offset is None:
            offset = 0
        where = query.where_clause or {}
        match = mongo_query_to_match_function(where)
        rows = [o for o in self.objects_as_list if match(o)]
        count = len(rows)
        if limit is None or limit < 0:
            limit = count
        # TODO: avoid recalculating
        returned_row = rows[offset : offset + limit]
        return QueryResult(query=query, num_rows=count, rows=returned_row)

    def query_facets(
        self, where: Dict = None, facet_columns: List[str] = None, facet_limit=DEFAULT_FACET_LIMIT, **kwargs
    ) -> Dict[str, Dict[str, int]]:
        match = mongo_query_to_match_function(where)
        rows = [o for o in self.objects_as_list if match(o)]
        if not facet_columns:
            facet_columns = self.class_definition().attributes.keys()
        facet_results = {c: {} for c in facet_columns}
        for row in rows:
            for fc in facet_columns:
                if fc in row:
                    v = row[fc]
                    if not isinstance(v, str):
                        v = str(v)
                    if v not in facet_results[fc]:
                        facet_results[fc][v] = 1
                    else:
                        facet_results[fc][v] += 1
        return {fc: list(facet_results[fc].items()) for fc in facet_results}



import pytest
from google.auth import default
from google.cloud import bigquery


@pytest.mark.skip("BigQuery tests require credentials")
def test_bq():
    """
    Tests behavior of BigQuery.

    To run this test:

    - ``gcloud auth application-default login``
    - place credentials in ``~/.config/gcloud/application_default_credentials.json``

    """
    credentials, project = default()
    client = bigquery.Client(credentials=credentials, project=project)

    # Construct a query
    query = """
        SELECT name, SUM(number) as total_people
        FROM `bigquery-public-data.usa_names.usa_1910_2013`
        WHERE state = 'TX'
        GROUP BY name
        ORDER BY total_people DESC
        LIMIT 20
    """

    # Run the query
    query_job = client.query(query)

    # Get results
    results = query_job.result()

    # Print results
    for row in results:
        print(f"{row.name}: {row.total_people}")


@pytest.mark.skip("BigQuery tests require credentials")
def test_sra():
    """
    Tests behavior of BigQuery.

    To run this test:

    - ``gcloud auth application-default login``
    - place credentials in ``~/.config/gcloud/application_default_credentials.json``

    """
    credentials, project = default()
    client = bigquery.Client(credentials=credentials, project=project)

    # Construct a query
    query = """
        SELECT *
        FROM `nih-sra-datastore.sra.metadata` as s
        WHERE organism = 'Syngnathus scovelli' and ( ('sex_calc', 'female') in 
         UNNEST(s.attributes) and ('dev_stage_sam', 'Adult') in UNNEST(s.attributes) ) limit 10
    """

    # Run the query
    query_job = client.query(query)

    # Get results
    results = query_job.result()

    # Print results
    for row in results:
        print(type(row))
        print(f"{row.organism}: {row.attributes}")
        for a in row.attributes:
            print(f"  {a['k']} == {a['v']}")



"""Tests for linkml-data-browser."""

from pathlib import Path

THIS_DIR = Path(__file__).parent
INPUT_DIR = THIS_DIR / "input"
OUTPUT_DIR = THIS_DIR / "output"

PERSONINFO_SCHEMA = INPUT_DIR / "nested2.schema.yaml"

COUNTRIES_DIR = INPUT_DIR / "countries"
COUNTRIES_CONFIG = COUNTRIES_DIR / "config.yaml"
COUNTRIES_SCHEMA = COUNTRIES_DIR / "countries.linkml.yaml"
COUNTRIES_DB = COUNTRIES_DIR / "countries.db"
COUNTRIES_DATA_JSONL = COUNTRIES_DIR / "countries.jsonl"



"""Configuration file for the Sphinx documentation builder."""

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
from datetime import date

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "linkml-store"
copyright = f"{date.today().year}, Author 1 <author@org.org>"
author = "Author 1 <author@org.org>"


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# from https://github.com/cthoyt/ontoportal-client/blob/9862e26f8e374c3aef8707e3d5d69526c4d0fcd5/docs/source/conf.py
# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# A list of prefixes that are ignored when creating the module index. (new in Sphinx 0.6)
modindex_common_prefix = ["linkml_store."]

extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx_rtd_theme",
    "sphinx_click",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "myst_parser",
    "nbsphinx",
]



# generate autosummary pages
autosummary_generate = True

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

templates_path = ["_templates"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_logo = 'https://linkml.io/uploads/linkml-logo_color.png'
html_favicon = 'https://linkml.io/uploads/linkml-logo_color-no-words.png'
html_static_path = ["_static"]

# https://stackoverflow.com/questions/5599254/how-to-use-sphinxs-autodoc-to-document-a-classs-init-self-method
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__'
}


# Add this to create a parallel hierarchy
#add_module_names = False
#python_use_unqualified_type_names = True

# Ensure full module paths are used as page names
#autosummary_generate = True
#autosummary_imported_members = True


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
if os.path.exists("logo.png"):
    html_logo = "logo.png"



"""Plotting utilities for embedding visualizations."""

import logging
from typing import Dict, List, Optional, Union, Literal, Tuple
import numpy as np
from dataclasses import dataclass, field
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

from linkml_store.utils.embedding_utils import EmbeddingData
from linkml_store.plotting.dimensionality_reduction import ReductionResult

logger = logging.getLogger(__name__)


@dataclass
class EmbeddingPlotConfig:
    """Configuration for embedding plots."""

    # Visual encoding
    color_field: Optional[str] = None
    shape_field: Optional[str] = "collection"
    size_field: Optional[str] = None
    hover_fields: List[str] = field(default_factory=list)

    # Plot styling
    title: str = "Embedding Visualization"
    width: int = 800
    height: int = 600
    point_size: int = 8
    opacity: float = 0.7

    # Color schemes
    color_discrete_map: Optional[Dict] = None
    color_continuous_scale: str = "Viridis"

    # Shape mapping
    shape_map: Optional[Dict] = None
    marker_symbols: List[str] = field(default_factory=lambda: [
        "circle", "square", "diamond", "cross", "x",
        "triangle-up", "triangle-down", "pentagon", "hexagon", "star"
    ])

    # Display options
    show_legend: bool = True
    show_axes: bool = True
    dark_mode: bool = False


def plot_embeddings(
    embedding_data: EmbeddingData,
    reduction_result: ReductionResult,
    config: Optional[EmbeddingPlotConfig] = None,
    output_file: Optional[str] = None
) -> go.Figure:
    """
    Create interactive plot of embeddings.

    Args:
        embedding_data: Embedding data with metadata
        reduction_result: Dimensionality reduction results
        config: Plot configuration
        output_file: Optional path to save HTML file

    Returns:
        Plotly figure object
    """
    if config is None:
        config = EmbeddingPlotConfig()

    # Prepare data for plotting
    plot_data = _prepare_plot_data(embedding_data, reduction_result, config)

    # Create figure
    fig = _create_scatter_plot(plot_data, config)

    # Apply styling
    fig = _style_figure(fig, config, reduction_result)

    # Save if requested
    if output_file:
        fig.write_html(output_file)
        logger.info(f"Saved plot to {output_file}")

    return fig


def _prepare_plot_data(
    embedding_data: EmbeddingData,
    reduction_result: ReductionResult,
    config: EmbeddingPlotConfig
) -> Dict:
    """Prepare data for plotting."""
    data = {
        "x": reduction_result.coordinates[:, 0],
        "y": reduction_result.coordinates[:, 1],
        "ids": embedding_data.object_ids,
        "collection": embedding_data.collection_names,
    }

    # Add metadata fields
    for field in config.hover_fields:
        values = embedding_data.get_metadata_values(field)
        if values:
            data[field] = values

    # Add color field
    if config.color_field:
        if config.color_field == "collection":
            data["color"] = embedding_data.collection_names
        else:
            data["color"] = embedding_data.get_metadata_values(config.color_field)

    # Add shape field
    if config.shape_field:
        if config.shape_field == "collection":
            data["shape"] = embedding_data.collection_names
        else:
            data["shape"] = embedding_data.get_metadata_values(config.shape_field)

    # Add size field
    if config.size_field:
        data["size"] = embedding_data.get_metadata_values(config.size_field)
    else:
        data["size"] = [config.point_size] * embedding_data.n_samples

    return data


def _create_scatter_plot(
    plot_data: Dict,
    config: EmbeddingPlotConfig
) -> go.Figure:
    """Create the scatter plot."""
    fig = go.Figure()

    # Determine if we need to create separate traces for different shapes
    if "shape" in plot_data and config.shape_field:
        unique_shapes = list(set(plot_data["shape"]))

        # Create shape mapping
        if config.shape_map:
            shape_map = config.shape_map
        else:
            shape_map = {
                shape: config.marker_symbols[i % len(config.marker_symbols)]
                for i, shape in enumerate(unique_shapes)
            }

        # Create trace for each shape category
        for shape_value in unique_shapes:
            mask = [s == shape_value for s in plot_data["shape"]]
            trace_data = {
                key: [v for i, v in enumerate(values) if mask[i]]
                for key, values in plot_data.items()
            }

            # Prepare hover text
            hover_text = _create_hover_text(trace_data, config)

            # Determine color
            if "color" in trace_data:
                color_values = trace_data["color"]
                # Check if categorical or continuous
                if color_values and all(isinstance(v, (int, float)) for v in color_values if v is not None):
                    marker_color = color_values
                    marker_colorscale = config.color_continuous_scale
                else:
                    # Categorical - don't use color map for plotly, let it auto-assign
                    # Just use the raw categorical values
                    marker_color = color_values
                    marker_colorscale = None
            else:
                marker_color = None
                marker_colorscale = None

            # When using separate traces, we can't use categorical colors directly
            # Use a single color per trace instead
            if marker_colorscale is None and marker_color is not None:
                # For categorical, just use the trace name for automatic coloring
                marker_dict = dict(
                    symbol=shape_map.get(shape_value, "circle"),
                    size=trace_data.get("size", config.point_size),
                    opacity=config.opacity,
                    line=dict(width=0.5, color="white")
                )
            else:
                # For continuous colors
                marker_dict = dict(
                    symbol=shape_map.get(shape_value, "circle"),
                    size=trace_data.get("size", config.point_size),
                    color=marker_color,
                    colorscale=marker_colorscale,
                    opacity=config.opacity,
                    line=dict(width=0.5, color="white")
                )

            trace = go.Scatter(
                x=trace_data["x"],
                y=trace_data["y"],
                mode="markers",
                name=str(shape_value),
                text=hover_text,
                hovertemplate="%{text}<extra></extra>",
                marker=marker_dict
            )
            fig.add_trace(trace)
    else:
        # Single trace for all points
        hover_text = _create_hover_text(plot_data, config)

        # Handle colors
        if "color" in plot_data:
            color_values = plot_data["color"]
            if all(isinstance(v, (int, float)) for v in color_values if v is not None):
                marker_color = color_values
                marker_colorscale = config.color_continuous_scale
                showscale = True
            else:
                if config.color_discrete_map:
                    marker_color = [config.color_discrete_map.get(c, c) for c in color_values]
                else:
                    marker_color = color_values
                marker_colorscale = None
                showscale = False
        else:
            marker_color = "blue"
            marker_colorscale = None
            showscale = False

        trace = go.Scatter(
            x=plot_data["x"],
            y=plot_data["y"],
            mode="markers",
            text=hover_text,
            hovertemplate="%{text}<extra></extra>",
            marker=dict(
                size=plot_data.get("size", config.point_size),
                color=marker_color,
                colorscale=marker_colorscale,
                showscale=showscale,
                opacity=config.opacity,
                line=dict(width=0.5, color="white")
            )
        )
        fig.add_trace(trace)

    return fig


def _create_hover_text(
    data: Dict,
    config: EmbeddingPlotConfig
) -> List[str]:
    """Create hover text for each point."""
    hover_texts = []
    n_points = len(data["x"])

    for i in range(n_points):
        lines = []

        # Add ID
        if "ids" in data:
            lines.append(f"<b>ID:</b> {data['ids'][i]}")

        # Add collection
        if "collection" in data:
            lines.append(f"<b>Collection:</b> {data['collection'][i]}")

        # Add hover fields
        for field in config.hover_fields:
            if field in data and data[field][i] is not None:
                lines.append(f"<b>{field}:</b> {data[field][i]}")

        # Add coordinates
        lines.append(f"<b>Coordinates:</b> ({data['x'][i]:.3f}, {data['y'][i]:.3f})")

        hover_texts.append("<br>".join(lines))

    return hover_texts


def _style_figure(
    fig: go.Figure,
    config: EmbeddingPlotConfig,
    reduction_result: ReductionResult
) -> go.Figure:
    """Apply styling to the figure."""
    # Create subtitle with method info
    subtitle = f"Method: {reduction_result.method.upper()}"
    if reduction_result.explained_variance:
        subtitle += f" | Explained variance: {reduction_result.explained_variance:.2%}"

    full_title = f"{config.title}<br><sub>{subtitle}</sub>"

    # Update layout
    layout_updates = dict(
        title=full_title,
        width=config.width,
        height=config.height,
        showlegend=config.show_legend,
        hovermode="closest",
        xaxis=dict(
            title="Component 1",
            showgrid=True,
            gridwidth=1,
            gridcolor="LightGray" if not config.dark_mode else "DarkGray",
            showline=config.show_axes,
            zeroline=True,
        ),
        yaxis=dict(
            title="Component 2",
            showgrid=True,
            gridwidth=1,
            gridcolor="LightGray" if not config.dark_mode else "DarkGray",
            showline=config.show_axes,
            zeroline=True,
        )
    )

    if config.dark_mode:
        layout_updates.update(
            template="plotly_dark",
            paper_bgcolor="black",
            plot_bgcolor="black",
        )
    else:
        layout_updates.update(
            template="plotly_white",
            paper_bgcolor="white",
            plot_bgcolor="white",
        )

    fig.update_layout(**layout_updates)

    return fig


def plot_embeddings_comparison(
    embedding_datasets: Dict[str, Tuple[EmbeddingData, ReductionResult]],
    config: Optional[EmbeddingPlotConfig] = None,
    output_file: Optional[str] = None
) -> go.Figure:
    """
    Create comparison plot of multiple embedding datasets.

    Args:
        embedding_datasets: Dictionary of (name -> (embedding_data, reduction_result))
        config: Plot configuration
        output_file: Optional path to save HTML file

    Returns:
        Plotly figure with subplots
    """
    if config is None:
        config = EmbeddingPlotConfig()

    n_datasets = len(embedding_datasets)
    n_cols = min(n_datasets, 3)
    n_rows = (n_datasets + n_cols - 1) // n_cols

    # Create subplots
    subplot_titles = list(embedding_datasets.keys())
    fig = make_subplots(
        rows=n_rows,
        cols=n_cols,
        subplot_titles=subplot_titles,
        horizontal_spacing=0.1,
        vertical_spacing=0.15
    )

    # Add each dataset as a subplot
    for idx, (name, (emb_data, red_result)) in enumerate(embedding_datasets.items()):
        row = idx // n_cols + 1
        col = idx % n_cols + 1

        plot_data = _prepare_plot_data(emb_data, red_result, config)

        # Create scatter trace
        hover_text = _create_hover_text(plot_data, config)

        trace = go.Scatter(
            x=plot_data["x"],
            y=plot_data["y"],
            mode="markers",
            name=name,
            text=hover_text,
            hovertemplate="%{text}<extra></extra>",
            marker=dict(
                size=plot_data.get("size", config.point_size),
                color=plot_data.get("color", "blue"),
                opacity=config.opacity,
            ),
            showlegend=(idx == 0)  # Only show legend for first subplot
        )

        fig.add_trace(trace, row=row, col=col)

        # Update axes labels
        fig.update_xaxes(title_text="Component 1", row=row, col=col)
        fig.update_yaxes(title_text="Component 2", row=row, col=col)

    # Update overall layout
    fig.update_layout(
        title=config.title,
        width=config.width * n_cols // 2,
        height=config.height * n_rows // 2,
        showlegend=config.show_legend,
        hovermode="closest"
    )

    # Save if requested
    if output_file:
        fig.write_html(output_file)
        logger.info(f"Saved comparison plot to {output_file}")

    return fig


def plot_embedding_clusters(
    embedding_data: EmbeddingData,
    reduction_result: ReductionResult,
    cluster_labels: np.ndarray,
    config: Optional[EmbeddingPlotConfig] = None,
    output_file: Optional[str] = None
) -> go.Figure:
    """
    Plot embeddings with cluster assignments.

    Args:
        embedding_data: Embedding data
        reduction_result: Dimensionality reduction results
        cluster_labels: Cluster labels for each point
        config: Plot configuration
        output_file: Optional path to save HTML file

    Returns:
        Plotly figure with clusters
    """
    if config is None:
        config = EmbeddingPlotConfig()

    # Override color field to use clusters
    config.color_field = "cluster"

    # Add cluster labels to metadata
    for i, label in enumerate(cluster_labels):
        if i < len(embedding_data.metadata):
            embedding_data.metadata[i]["cluster"] = f"Cluster {label}"

    # Create plot
    fig = plot_embeddings(embedding_data, reduction_result, config)

    # Add cluster centroids if possible
    unique_labels = np.unique(cluster_labels)
    if len(unique_labels) < 50:  # Only show centroids for reasonable number of clusters
        for label in unique_labels:
            mask = cluster_labels == label
            centroid_x = np.mean(reduction_result.coordinates[mask, 0])
            centroid_y = np.mean(reduction_result.coordinates[mask, 1])

            fig.add_trace(go.Scatter(
                x=[centroid_x],
                y=[centroid_y],
                mode="markers+text",
                marker=dict(
                    size=15,
                    symbol="star",
                    color="black",
                    line=dict(width=2, color="white")
                ),
                text=f"C{label}",
                textposition="top center",
                showlegend=False,
                hovertemplate=f"Cluster {label} centroid<extra></extra>"
            ))

    # Update title
    fig.update_layout(
        title=f"{config.title}<br><sub>Clusters: {len(unique_labels)}</sub>"
    )

    # Save if requested
    if output_file:
        fig.write_html(output_file)
        logger.info(f"Saved cluster plot to {output_file}")

    return fig


import logging
from typing import Any, Optional, Tuple, Type, Union

import sqlalchemy
import sqlalchemy.sql.sqltypes as sqlt
from linkml_runtime.linkml_model import SchemaDefinition, SlotDefinition
from linkml_runtime.utils.schema_builder import SchemaBuilder
from sqlalchemy import MetaData, quoted_name

from linkml_store.api.queries import Query

logger = logging.getLogger(__name__)

TYPE_MAP = {
    sqlt.TEXT: "string",
    sqlt.INTEGER: "integer",
    sqlt.FLOAT: "float",
}

OP_MAP = {
    "eq": "=",
    "$in": "ARRAY_CONTAINS",  ## mongodb
    "$contains": "ARRAY_CONTAINS",  ## TODO: this is chromadb-specific
    "in": "ARRAY_CONTAINS",
}


def _map_type(typ: Type) -> str:
    for k, v in TYPE_MAP.items():
        if isinstance(typ, k):
            return v
    return "string"


def where_clause_to_sql(query: Query) -> str:
    if not query.where_clause:
        return ""
    if isinstance(query.where_clause, str):
        where_clause_sql = query.where_clause
    elif isinstance(query.where_clause, list):
        where_clause_sql = " AND ".join(query.where_clause)
    elif isinstance(query.where_clause, dict):
        conjs = []
        for k, v in query.where_clause.items():
            conjs.extend(col_val_constraints_to_conjs(k, v))
        where_clause_sql = " AND ".join(conjs)

    else:
        raise ValueError(f"Invalid where_clause type: {type(query.where_clause)}")
    return "WHERE " + where_clause_sql


def col_val_constraints_to_conjs(col_name: str, val_constraints: Any) -> list:
    if val_constraints is None:
        return []

    def _quote(v: Any):
        if isinstance(v, str):
            # escape internal vs
            v = v.replace("'", "''")
            return f"'{v}'"
        else:
            return v

    if isinstance(val_constraints, dict):
        conjs = []
        for k, v in val_constraints.items():
            if k in OP_MAP:
                if k == "$in" and isinstance(v, list):
                    v_mapped = [_quote(v1) for v1 in v]
                    t = f"{col_name} IN ({', '.join(v_mapped)})"
                else:
                    t = f"{OP_MAP[k]}({col_name}, {_quote(v)})"
            else:
                t = f"{col_name} {k} {_quote(v)}"
            conjs.append(t)
        return conjs
    else:
        return [f"{col_name} = {_quote(val_constraints)}"]


def query_to_sql(query: Query, count=False, limit=None, offset: Optional[int] = None):
    select_cols = query.select_cols if query.select_cols else ["*"]
    if count:
        sql_str = ["SELECT COUNT(*)"]
    else:
        sql_str = [f"SELECT {', '.join(select_cols)}"]
    sql_str.append(f"FROM {query.from_table}")
    sql_str.append(where_clause_to_sql(query))
    if not count:
        if query.sort_by:
            sql_str.append(f"ORDER BY {', '.join(query.sort_by)}")
    if not count:
        if limit is None:
            limit = query.limit
        if limit is None:
            limit = 100
        if limit < 0:
            limit = None
        if limit is not None:
            sql_str.append(f" LIMIT {limit}")
        offset = offset if offset else query.offset
        if offset:
            sql_str.append(f" OFFSET {offset}")
    sql_str = [line for line in sql_str if line]
    return "\n".join(sql_str)


def facet_count_sql(query: Query, facet_column: Union[str, Tuple[str, ...]], multivalued=False, limit=100) -> str:
    # Create a modified WHERE clause that excludes conditions directly related to facet_column
    modified_where = None
    if query.where_clause:
        where_clause_sql = where_clause_to_sql(query)
        # Split the where clause into conditions and exclude those related to the facet_column
        conditions = [cond for cond in where_clause_sql.split(" AND ") if not cond.startswith(f"{facet_column} ")]
        modified_where = " AND ".join(conditions)

    def make_col_safe(col):
        return '"' + quoted_name(col, True) + '"' if " " in col else col

    if isinstance(facet_column, str):
        facet_column = make_col_safe(facet_column)
    if isinstance(facet_column, tuple):
        facet_column = [make_col_safe(col) for col in facet_column]
        if multivalued:
            raise NotImplementedError("Multivalued facets are not supported for multiple columns")
        facet_column = ", ".join(facet_column)
    from_table = query.from_table
    if multivalued:
        from_table = f"(SELECT UNNEST({facet_column}) as {facet_column} FROM {query.from_table}"
        from_table += f" {modified_where}" if modified_where else ""
        from_table += ")"
    else:
        from_table += f" {modified_where}" if modified_where else ""
    sql_str = [f"SELECT {facet_column}, COUNT(*) as count", f"FROM {from_table}"]
    # if modified_where:
    #    sql_str.append(f"{modified_where}")
    sql_str.append(f"GROUP BY {facet_column}")
    sql_str.append("ORDER BY count DESC")  # Optional, order by count for convenience
    if limit is not None:
        sql_str.append(f"LIMIT {limit}")
    return "\n".join(sql_str)


def introspect_schema(engine: sqlalchemy.Engine) -> SchemaDefinition:
    """
    Introspect a database schema and return a SchemaDefinition object

    :param engine:
    :return:
    """
    metadata_obj = MetaData()
    logging.info(f"Reflecting using {engine}")
    metadata_obj.reflect(bind=engine)
    sb = SchemaBuilder()
    schema = sb.schema
    for table in metadata_obj.sorted_tables:
        logging.info(f"Importing {table.name}")
        sb.add_class(table.name)
        cls = schema.classes[table.name]
        pks = [column for column in table.columns if column.primary_key]
        if len(pks) == 1:
            pk = pks.pop().name
        else:
            pk = None
        for column in table.columns:
            slot = SlotDefinition(column.name)
            cls.attributes[slot.name] = slot
            if pk and pk == column.name:
                slot.identifier = True
            if column.foreign_keys:
                for fk in column.foreign_keys:
                    [fk_table, fk_table_col] = str(fk.column).split(".")
                    slot.range = fk_table
            else:
                slot.range = _map_type(column.type)
    return schema



import csv
import gzip
import hashlib
import io
import json
import logging
import re
import sys
import tarfile
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, TextIO, Type, Union

import pandas as pd
import pystow
import xmltodict
import yaml
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class Format(Enum):
    """
    Supported generic file formats for loading and rendering objects.
    """

    JSON = "json"
    JSONL = "jsonl"
    YAML = "yaml"
    YAMLL = "yamll"
    TOML = "toml"
    TSV = "tsv"
    CSV = "csv"
    XML = "xml"
    TURTLE = "turtle"
    RDFXML = "rdfxml"
    TEXT = "text"
    TEXTLINES = "textlines"
    OBO = "obo"
    FASTA = "fasta"
    GMT = "gmt"
    DAT = "dat"
    MARKDOWN = "markdown"
    PKL = "pkl"
    RDS = "rds"
    PYTHON = "python"
    PARQUET = "parquet"
    HDF5 = "hdf5"
    NETCDF = "netcdf"
    FORMATTED = "formatted"
    TABLE = "table"
    XLSX = "xlsx"
    PNG = "png"
    SQLDUMP_DUCKDB = "duckdb"
    SQLDUMP_POSTGRES = "postgres"
    DUMP_MONGODB = "mongodb"

    @classmethod
    def guess_format(cls, file_name: str) -> Optional["Format"]:
        ext = Path(file_name).suffix.lower()

        format_map = {
            ".json": cls.JSON,
            ".jsonl": cls.JSONL,
            ".yaml": cls.YAML,
            ".yml": cls.YAML,
            ".yamll": cls.YAMLL,
            ".tsv": cls.TSV,
            ".csv": cls.CSV,
            ".txt": cls.TEXT,
            ".xml": cls.XML,
            ".owx": cls.XML,
            ".owl": cls.RDFXML,
            ".ttl": cls.TURTLE,
            ".md": cls.MARKDOWN,
            ".py": cls.PYTHON,
            ".parquet": cls.PARQUET,
            ".pq": cls.PARQUET,
        }
        fmt = format_map.get(ext, None)
        if fmt is None:
            if ext.startswith("."):
                ext = ext[1:]
            if ext in [f.value for f in Format]:
                return Format(ext)
        return fmt

    def is_dump_format(self):
        return self in [Format.SQLDUMP_DUCKDB, Format.SQLDUMP_POSTGRES, Format.DUMP_MONGODB]

    def is_binary_format(self):
        return self in [Format.PARQUET, Format.XLSX]

    def is_xsv(self):
        return self in [Format.TSV, Format.CSV]


def load_objects_from_url(
    url: str,
    format: Union[Format, str] = None,
    expected_type: Type = None,
    local_path: Optional[str] = None,
    **kwargs,
) -> List[Dict[str, Any]]:
    """
    Load objects from a URL in JSON, JSONLines, YAML, CSV, or TSV format.

    :param url: The URL to the file.
    :param format: The format of the file. Can be a Format enum or a string value.
    :param expected_type: The target type to load the objects into.
    :param local_path: The local path to save the file to.
    :return: A list of dictionaries representing the loaded objects.
    """
    local_path = pystow.ensure("linkml", "linkml-store", url=url)
    logger.info(f"synced to {local_path}")
    objs = load_objects(local_path, format=format, expected_type=expected_type, **kwargs)
    if not objs:
        raise ValueError(f"No objects loaded from URL: {url}")
    return objs


def clean_pandas_value(v):
    """Clean a single value from pandas."""
    import math

    if isinstance(v, float):
        if math.isnan(v) or math.isinf(v):
            return None
        return float(v)  # Ensures proper float type
    return v


def clean_nested_structure(obj):
    """Recursively clean a nested structure of dicts/lists from pandas."""
    if isinstance(obj, dict):
        return {k: clean_nested_structure(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [clean_nested_structure(item) for item in obj]  # Fixed: using 'item' instead of 'v'
    else:
        return clean_pandas_value(obj)


def process_file(
    f: IO,
    format: Format,
    expected_type: Optional[Type] = None,
    header_comment_token: Optional[str] = None,
    format_options: Optional[Dict[str, Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Process a single file and return a list of objects.

    :param f: The file object.
    :param format: The format of the file.
    :param expected_type: The expected type of the objects.
    :param header_comment_token: Token used for header comments to be skipped
    :return:
    """
    if format_options is None:
        format_options = {}
    if format == Format.YAMLL:
        format = Format.YAML
        expected_type = list
    if format == Format.JSON:
        objs = json.load(f)
    elif format == Format.JSONL:
        objs = [json.loads(line) for line in f]
    elif format == Format.YAML:
        if expected_type and expected_type == list:  # noqa E721
            objs = list(yaml.safe_load_all(f))
            # allow YAML with a `---` with no object before it
            objs = [obj for obj in objs if obj is not None]
        else:
            objs = yaml.safe_load(f)
    elif format == Format.TOML:
        import toml

        objs = toml.load(f)
        if not isinstance(objs, list):
            objs = [objs]
    elif format == Format.TEXTLINES:
        objs = f.readlines()
    elif format in [Format.TSV, Format.CSV]:
        if header_comment_token:
            while True:
                pos = f.tell()
                line = f.readline()
                if not line.startswith(header_comment_token):
                    f.seek(pos)
                    break
        delimiter = "\t" if format == Format.TSV else ","
        reader = csv.DictReader(f, delimiter=delimiter)
        objs = list(reader)
    elif format == Format.XML:
        objs = xmltodict.parse(f.read())
    elif format == Format.PKL:
        objs = pd.read_pickle(f).to_dict(orient="records")
    elif format == Format.RDS:
        import pyreadr
        objs = pyreadr.read_r(f)
    elif format == Format.XLSX:
        xls = pd.ExcelFile(f)
        objs = {sheet: clean_nested_structure(xls.parse(sheet).to_dict(orient="records")) for sheet in xls.sheet_names}
    elif format == Format.TEXT:
        txt = f.read()
        objs = [
            {
                "name": Path(f.name).name,
                "path": f.name,
                "content": txt,
                "size": len(txt),
                "lines": txt.count("\n") + 1,
                "md5": hashlib.md5(txt.encode()).hexdigest(),
            }
        ]
    elif format == Format.GMT:
        objs = []
        lib_name = Path(f.name).name
        for line in f:
            parts = line.strip().split("\t")
            desc = parts[1]
            objs.append(
                {
                    "library": lib_name,
                    "uid": f"{lib_name}.{parts[0]}",
                    "name": parts[0],
                    "description": desc if desc else None,
                    "genes": parts[2:],
                }
            )
    elif format == Format.FASTA:
        objs = []
        current_obj = None
        for line in f:
            line = line.strip()
            if line.startswith(">"):
                if current_obj:
                    objs.append(current_obj)
                current_obj = {"id": line[1:], "sequence": ""}
            else:
                current_obj["sequence"] += line
        if current_obj:
            objs.append(current_obj)
    elif format == Format.OBO:
        blocks = split_document(f.read(), "\n\n")
        id_pattern = re.compile(r"id: (\S+)")

        def get_id(block):
            m = id_pattern.search(block)
            return m.group(1) if m else None

        objs = [{"id": get_id(block), "content": block} for block in blocks]
        objs = [obj for obj in objs if obj["id"]]
    elif format == Format.DAT:
        from linkml_store.utils.dat_parser import parse_sib_format

        _, objs = parse_sib_format(f.read())
    elif format in (Format.RDFXML, Format.TURTLE):
        import lightrdf

        parser = lightrdf.Parser()
        objs = []
        ext_fmt = "rdfxml"
        if format == Format.TURTLE:
            ext_fmt = "ttl"
        bytesio = io.BytesIO(f.read().encode("utf-8"))
        buffer = io.BufferedReader(bytesio)
        for s, p, o in parser.parse(buffer, base_iri=None, format=ext_fmt):
            obj = {
                "subject": s,
                "predicate": p,
                "object": o,
            }
            if format_options.get("pivot", False):
                obj = {
                    "subject": s,
                    p: o,
                }
            objs.append(obj)
    elif format == Format.PARQUET:
        import pyarrow.parquet as pq

        table = pq.read_table(f)
        objs = table.to_pandas().to_dict(orient="records")
    elif format in [Format.PYTHON, Format.FORMATTED, Format.TABLE]:
        raise ValueError(f"Format {format} is not supported for loading objects")
    else:
        raise ValueError(f"Unsupported file format: {format}")

    if not isinstance(objs, list):
        objs = [objs]
    return objs


def load_objects(
    file_path: Union[str, Path],
    format: Optional[Union[Format, str]] = None,
    compression: Optional[str] = None,
    expected_type: Optional[Type] = None,
    header_comment_token: Optional[str] = None,
    select_query: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Load objects from a file or archive in supported formats.
    For tgz archives, it processes all files and concatenates the results.

    TODO: Add schema hints for CSV/TSV parsing.

    :param file_path: The path to the file or archive.
    :param format: The format of the file. Can be a Format enum or a string value.
    :param compression: The compression type. Supports 'gz' for gzip and 'tgz' for tar.gz.
    :param expected_type: The target type to load the objects into, e.g. list
    :param header_comment_token: Token used for header comments to be skipped
    :param select_query: JSONPath query to select specific objects from the loaded data.
    :return: A list of dictionaries representing the loaded objects.
    """
    if isinstance(file_path, Path):
        file_path = str(file_path)

    for url_scheme in ["http", "https", "ftp"]:
        if file_path.startswith(f"{url_scheme}://"):
            return load_objects_from_url(
                file_path,
                format=format,
                expected_type=expected_type,
            )

    if isinstance(format, str):
        format = Format(format)

    all_objects = []

    if compression == "tgz":
        with tarfile.open(file_path, "r:gz") as tar:
            for member in tar.getmembers():
                if member.isfile():
                    f = tar.extractfile(member)
                    if f:
                        content = io.TextIOWrapper(f)
                        member_format = Format.guess_format(member.name) if not format else format
                        logger.debug(f"Processing tar member {member.name} with format {member_format}")
                        all_objects.extend(process_file(content, member_format, expected_type, header_comment_token))
    else:
        if Path(file_path).is_dir():
            raise ValueError(f"{file_path} is a dir, which is invalid for {format}")
        open_func = gzip.open if compression == "gz" else open
        format = Format.guess_format(file_path) if not format else format
        mode = "rb" if (format and format.is_binary_format()) or compression == "gz" else "r"
        with open_func(file_path, mode) if file_path != "-" else sys.stdin as f:
            if compression == "gz" and mode == "r":
                f = io.TextIOWrapper(f)
            all_objects = process_file(f, format, expected_type, header_comment_token)

    logger.debug(f"Loaded {len(all_objects)} objects from {file_path}")
    all_objects = transform_objects(all_objects, select_query)
    return all_objects

def transform_objects(all_objects: List[Dict[str, Any]], select_query: Optional[str]) -> List[Dict[str, Any]]:
    if not select_query:
        return all_objects
    import jsonpath_ng as jp

    path_expr = jp.parse(select_query)
    new_objs = []
    for obj in all_objects:
        for match in path_expr.find(obj):
            logging.debug(f"Match: {match.value}")
            if isinstance(match.value, list):
                new_objs.extend(match.value)
            else:
                new_objs.append(match.value)
    all_objects = new_objs
    return all_objects

def remove_control_chars_from_df(df: pd.DataFrame) -> pd.DataFrame:
    df_clean = df.copy()
    for col in df_clean.select_dtypes(include=['object']).columns:
        df_clean[col] = df_clean[col].astype(str).str.replace(r'[\x00-\x1f\x7f-\x9f]', '', regex=True)
    return df_clean

def write_output(
    data: Union[List[Dict[str, Any]], Dict[str, Any], pd.DataFrame],
    format: Union[Format, str] = Format.YAML,
    target: Optional[Union[TextIO, str, Path]] = None,
) -> None:
    """
    Write output data to a file in JSON, JSONLines, YAML, CSV, or TSV format.

    >>> write_output([{"a": 1, "b": 2}, {"a": 3, "b": 4}], Format.JSON, sys.stdout)
    [
      {
        "a": 1,
        "b": 2
      },
      {
        "a": 3,
        "b": 4
        }
    ]
    """
    if isinstance(format, str):
        format = Format(format)
    if format == Format.XLSX:
        if not target:
            raise ValueError("XLSX output requires a target file")
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)
        data = remove_control_chars_from_df(data)
        data.to_excel(target, index=False)
        return
    output_str = render_output(data, format)
    if target:
        if isinstance(target, str):
            with open(target, "w") as target:
                target.write(output_str)
        else:
            target.write(output_str)
    else:
        print(output_str)


def render_output(
    data: Union[List[Dict[str, Any]], Dict[str, Any], pd.DataFrame, List[BaseModel]],
    format: Optional[Union[Format, str]] = Format.YAML,
) -> str:
    """
    Render output data in JSON, JSONLines, YAML, CSV, or TSV format.

    >>> print(render_output([{"a": 1, "b": 2}, {"a": 3, "b": 4}], Format.JSON))
    [
      {
        "a": 1,
        "b": 2
      },
      {
        "a": 3,
        "b": 4
      }
    ]


    :param data: The data to be rendered.
    :param format: The desired output format. Can be a Format enum or a string value.
    :return: The rendered output as a string.
    """
    if isinstance(format, str):
        format = Format(format)

    if format in (Format.FORMATTED, ):
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)
        return data.to_string(max_rows=None)

    if isinstance(data, pd.DataFrame):
        data = data.to_dict(orient="records")

    if isinstance(data, BaseModel):
        data = data.model_dump()

    if data and isinstance(data, list) and isinstance(data[0], BaseModel):
        data = [d.model_dump() if isinstance(d, BaseModel) else d for d in data]

    if isinstance(data, dict) and format in [Format.TSV, Format.CSV]:
        data = [data]

    if isinstance(data, BaseModel):
        data = data.model_dump()

    if format == Format.JSON:
        return json.dumps(data, indent=2, default=str)
    elif format == Format.JSONL:
        return "\n".join(json.dumps(obj) for obj in data)
    elif format == Format.PYTHON:
        return str(data)
    elif format == Format.MARKDOWN:

        def as_markdown(obj: dict):
            return "## Object\n\n" + "\n".join([f" * {k}: {v}" for k, v in obj.items()])

        return "\n\n".join([as_markdown(obj) for obj in data]) if isinstance(data, list) else as_markdown(data)
    elif format == Format.TABLE:
        from tabulate import tabulate

        return tabulate(pd.DataFrame(data), headers="keys", tablefmt="psql")
    elif format == Format.YAML:
        if isinstance(data, list):
            return yaml.safe_dump_all(data, sort_keys=False)
        else:
            return yaml.safe_dump(data, sort_keys=False)
    elif format == Format.TSV:
        output = StringIO()
        writer = csv.DictWriter(output, fieldnames=get_fieldnames(data), delimiter="\t")
        writer.writeheader()
        writer.writerows(data)
        return output.getvalue()
    elif format == Format.CSV:
        output = StringIO()
        writer = csv.DictWriter(output, fieldnames=get_fieldnames(data))
        writer.writeheader()
        writer.writerows(data)
        return output.getvalue()
    else:
        raise ValueError(f"Unsupported output format: {format}")


def get_fieldnames(data: List[Dict[str, Any]]) -> List[str]:
    """
    Get the fieldnames of a list of dictionaries.

    >>> get_fieldnames([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    ['a', 'b']

    :param data: The list of dictionaries.
    :return: The fieldnames.
    """
    fieldnames = []
    for obj in data:
        fieldnames.extend([k for k in obj.keys() if k not in fieldnames])
    return fieldnames


def guess_format(path: str) -> Optional[Format]:
    """
    Guess the format of a file based on its extension.

    >>> guess_format("data.json")
    <Format.JSON: 'json'>
    >>> guess_format("data.jsonl")
    <Format.JSONL: 'jsonl'>
    >>> guess_format("data.yaml")
    <Format.YAML: 'yaml'>
    >>> assert not guess_format("data")

    :param path: The path to the file.
    :return: The guessed format.
    """
    return Format.guess_format(path)


def split_document(doc: str, delimiter: str):
    """
    Split a document into parts based on a delimiter.

    :param doc: The document to split.
    :param delimiter: The delimiter.
    :return: The parts of the document.
    """
    return doc.split(delimiter)



import logging
import os
from typing import Any, Dict

import numpy as np
import pandas as pd
import streamlit as st
import yaml
from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition

from linkml_store import Client
from linkml_store.api import Collection
from linkml_store.api.queries import QueryResult

logger = logging.getLogger(__name__)

# Set page config to make layout "wide" by default
st.set_page_config(layout="wide")

config = None
if os.environ.get("LINKML_STORE_CONFIG"):
    with open(os.environ["LINKML_STORE_CONFIG"], "r") as f:
        config = yaml.safe_load(f)

# Initialize client
client = Client().from_config(config) if config else Client()

DEFAULT_LIMIT = 25

TABLES = ["gaf_association", "gaf_association_plus_violations_m", "gpi"]
DBS = {
    "mgi": TABLES,
    "goa_human": TABLES,
    "gcrp": TABLES,
}

DB_PATH = "db/{db}.db"


def init_reset_filters(cd: ClassDefinition, reset=False):
    for att_name in cd.attributes:
        key = f"filter_{att_name}"
        if reset or key not in st.session_state:
            st.session_state[key] = ""  # Assuming text input, adjust for other types


def apply_filters(collection: Collection, filters: Dict[str, Any], offset: int, limit: int, **kwargs) -> QueryResult:
    print(f"FILTERS={filters} // offset={offset}")
    qr = collection.find(filters, offset=offset, limit=limit, **kwargs)
    print(f"QR={qr.num_rows}")
    return qr


def render_filter_widget(collection: Collection, attribute: SlotDefinition):
    """Render appropriate Streamlit widget based on column type."""
    logger.info(f"Rendering filter widget: {attribute.name}")
    # print(f"{attribute.name} // RANGE={attribute.range}")
    # col_type = attribute.range
    col_name = attribute.name
    cols = st.sidebar.columns([1, 10])
    with cols[0]:
        if st.button("▼", key=f"facet_button_{col_name}", use_container_width=True):
            # Toggle the facet view state for this filter
            key = f"facet_view_{col_name}"
            if key not in st.session_state:
                st.session_state[key] = True
            else:
                st.session_state[key] = not st.session_state[key]
    with cols[1]:
        filter_value = st.text_input(f"Filter by {col_name}", key=f"filter_{col_name}")
    return filter_value
    # return st.sidebar.text_input(f"Filter by {col_name}")
    # elif col_type == "integer":
    #     max_value = con.execute(f"SELECT MAX({col_name}) FROM {tbl_name}").fetchall()[0][0]
    #     min_value = con.execute(f"SELECT MIN({col_name}) FROM {tbl_name}").fetchall()[0][0]
    #     return st.sidebar.slider(f"Filter by {col_name}", min_value, max_value, (min_value, max_value))
    # elif col_type in ['float', 'double']:
    #     max_value = con.execute(f"SELECT MAX({col_name}) FROM {tbl_name}").fetchall()[0][0]
    #     min_value = con.execute(f"SELECT MIN({col_name}) FROM {tbl_name}").fetchall()[0][0]
    #     return st.sidebar.slider(f"Filter by {col_name}", float(min_value), float(max_value),
    #                              (float(min_value), float(max_value)))
    # Add more data types as needed


# Main function to render the app
def main():
    st.title("LinkML Table Browser")
    db_names = list(client.databases.keys())
    selected_db = st.selectbox("Select a Database", db_names, key="db_selector")
    print(f"DB SELECTED={selected_db}")
    # con = duckdb.connect(DB_PATH.format(db=selected_db))
    # db_name = DB_PATH.format(db=selected_db)
    # database = DuckDBDatabase(f"duckdb:///{db_name}")
    database = client.get_database(selected_db)
    st.write(f"Connected to {selected_db}")
    candidate_tables = database.list_collection_names()
    print(f"COLLECtiONS={candidate_tables}")
    if len(candidate_tables) > 1:
        curr_table = st.selectbox("Select a Table", candidate_tables, key="table_selector")
    else:
        curr_table = candidate_tables[0]
    collection = database.get_collection(curr_table)
    print(f"CURR={collection.alias} // {collection.target_class_name}")
    cd = collection.class_definition()
    print(f"CD={cd.name} // {len(cd.attributes)}")
    filters = {}

    # Pagination setup
    session_state = st.session_state
    if "current_page" not in session_state:
        print(f"RESETTING CP// {session_state}")
        session_state.current_page = 0  # Start with page 0
    rows_per_page = DEFAULT_LIMIT

    init_reset_filters(cd)

    # Track if any filter has changed
    filter_changed = False
    for att in cd.attributes.values():
        att_name = att.name
        key = f"filter_{att_name}"
        prev_value = st.session_state[key]
        filter_widget = render_filter_widget(collection, att)
        if filter_widget is not None and filter_widget != "":
            filters[att_name] = filter_widget
        new_value = filters.get(att_name)
        if prev_value != new_value and not (not prev_value and not new_value):
            print(f"CHANGE FOR {att_name}: {prev_value} -> {new_value}")
            filter_changed = True
            # st.session_state[key] = new_value
        facet_key = f"facet_view_{att_name}"
        if facet_key in st.session_state and st.session_state[facet_key]:
            facet_results = collection.query_facets(filters, facet_columns=[att_name])
            facet_df = pd.DataFrame(facet_results[att_name])
            st.sidebar.write(facet_df)
    # If any filter has changed, reset pagination
    if filter_changed:
        print(f"FILTER CHANGED={filter_changed}")
        st.session_state.current_page = 0  # Reset offset
    result = apply_filters(collection, filters, session_state.current_page * rows_per_page, rows_per_page)
    # if filter_changed:
    #    facet_results = collection.query_facets(filters, facet_columns=["evidence_type"])
    #    print(f"FACET={facet_results}")
    st.write(f"Number of rows: {result.num_rows}")
    st.write(f"Page: {session_state.current_page + 1}")
    print(f"SESSION STATE: {session_state}")
    filtered_data = pd.DataFrame(result.rows)

    # Pagination buttons
    cols = st.columns(4)
    prev_button, next_button, _first_button, _last_button = cols[0], cols[1], cols[2], cols[3]

    if prev_button.button("Previous"):
        if session_state.current_page > 0:
            session_state.current_page -= 1
    if next_button.button("Next"):
        print(f"NEXT: CP={session_state.current_page} RPP={rows_per_page} NR={result.num_rows}")
        # Assuming result.num_rows gives the total number of rows after filtering, not just this page's rows
        if (session_state.current_page + 1) * rows_per_page < result.num_rows:
            session_state.current_page += 1
    if prev_button.button("First"):
        session_state.current_page = 0
    if prev_button.button("Last"):
        session_state.current_page = int(result.num_rows / rows_per_page)

    # Refresh the data after pagination change
    if "current_page" in session_state:
        result = apply_filters(collection, filters, session_state.current_page * rows_per_page, rows_per_page)
        filtered_data = pd.DataFrame(result.rows)

    # Add 'id' column to the DataFrame for easier tracking of changes; incremental
    filtered_data["id"] = range(1, len(filtered_data) + 1)
    edited_df = st.data_editor(filtered_data, width=2000, num_rows="dynamic")
    original_data = filtered_data

    for index, row in edited_df.iterrows():
        if np.isnan(row["id"]):
            print(f"INSERT: {row}")

    original_ids = set(original_data["id"])
    edited_ids = set(edited_df["id"])

    deleted_ids = original_ids - edited_ids
    for deleted_id in deleted_ids:
        print(f"DELETE: {original_data[original_data['id'] == deleted_id]}")

    # For modified rows, compare values row by row where ids match
    _modified_ids = []
    for id_val in original_ids & edited_ids:  # Intersection: ids present in both
        _original_row = original_data[original_data["id"] == id_val]
        _edited_row = edited_df[edited_df["id"] == id_val]
        # Assuming you have a function to compare rows and decide if they are different
        # if not rows_are_equal(original_row, edited_row):
        #    modified_ids.append(id_val)
        #    print(f"MODIFY: {original_row} -> {edited_row}")


if __name__ == "__main__":
    main()



import logging
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from pydantic import BaseModel

from linkml_store.utils.vector_utils import mmr_diversified_search, pairwise_cosine_similarity

INDEX_ITEM = np.ndarray

logger = logging.getLogger(__name__)


class TemplateSyntaxEnum(str, Enum):
    """
    Template syntax types.
    """

    jinja2 = "jinja2"
    fstring = "fstring"


class Indexer(BaseModel):
    """
    An indexer operates on a collection in order to search for objects.

    You should use a subcllass of this; this can be looked up dynqamically:

    >>> from linkml_store.index import get_indexer
    >>> indexer = get_indexer("simple")

    You can customize how objects are indexed by passing in a text template.
    For example, if your collection has objects with "name" and "profession" attributes,
    you can index them as "{name} {profession}".

    >>> indexer = get_indexer("simple", text_template="{name} :: {profession}")

    By default, python fstrings are assumed.

    We can test this works using the :ref:`object_to_text` method (normally
    you would never need to call this directly, but it's useful for testing):

    >>> obj = {"name": "John", "profession": "doctor"}
    >>> indexer.object_to_text(obj)
    'John :: doctor'

    You can also use Jinja2 templates; this gives more flexibility and logic,
    e.g. conditional formatting:

    >>> tmpl = "{{name}}{% if profession %} :: {{profession}}{% endif %}"
    >>> indexer = get_indexer("simple", text_template=tmpl, text_template_syntax=TemplateSyntaxEnum.jinja2)
    >>> indexer.object_to_text(obj)
    'John :: doctor'
    >>> indexer.object_to_text({"name": "John"})
    'John'

    You can also specify which attributes to index:

    >>> indexer = get_indexer("simple", index_attributes=["name"])
    >>> indexer.object_to_text(obj)
    'John'

    The purpose of an indexer is to translate a collection of objects into a collection of objects
    such as vectors for purposes such as search. Unless you are implementing your own indexer, you
    generally don't need to use the methods that return vectors, but we can examine their behavior
    to get a sense of how they work.

    >>> vectors = indexer.objects_to_vectors([{"name": "Aardvark"}, {"name": "Aardwolf"}, {"name": "Zesty"}])
    >>> assert pairwise_cosine_similarity(vectors[0], vectors[1]) > pairwise_cosine_similarity(vectors[0], vectors[2])

    Note you should consult the documentation for the specific indexer you are using for more details on
    how text is converted to vectors.

    """

    name: Optional[str] = None
    index_type: Optional[str] = None
    index_function: Optional[Callable] = None
    distance_function: Optional[Callable] = None
    index_attributes: Optional[List[str]] = None
    text_template: Optional[str] = None
    text_template_syntax: Optional[TemplateSyntaxEnum] = None
    filter_nulls: Optional[bool] = True
    vector_default_length: Optional[int] = 1000
    index_field: Optional[str] = "__index__"
    index_value_field: Optional[str] = "__index_value__"

    def object_to_vector(self, obj: Dict[str, Any]) -> INDEX_ITEM:
        """
        Convert an object to an indexable object

        :param obj:
        :return:
        """
        return self.text_to_vector(self.object_to_text(obj))

    def objects_to_vectors(self, objs: List[Dict[str, Any]]) -> List[INDEX_ITEM]:
        """
        Convert a list of objects to indexable objects

        :param objs:
        :return: list of vectors
        """
        return self.texts_to_vectors([self.object_to_text(obj) for obj in objs])

    def texts_to_vectors(self, texts: List[str], cache: bool = None, **kwargs) -> List[INDEX_ITEM]:
        """
        Convert a list of texts to indexable objects

        :param texts:
        :return:
        """
        return [self.text_to_vector(text, cache=cache, **kwargs) for text in texts]

    def text_to_vector(self, text: str, cache: bool = None, **kwargs) -> INDEX_ITEM:
        """
        Convert a text to an indexable object

        :param text:
        :param cache:
        :return:
        """
        raise NotImplementedError

    def object_to_text(self, obj: Dict[str, Any]) -> str:
        """
        Convert an object to a text representation

        :param obj:
        :return:
        """
        if self.index_attributes:
            if len(self.index_attributes) == 1 and not self.text_template:
                return str(obj[self.index_attributes[0]])
            obj = {k: v for k, v in obj.items() if k in self.index_attributes}
        if self.filter_nulls:
            obj = {k: v for k, v in obj.items() if v is not None}
        if self.text_template:
            syntax = self.text_template_syntax
            if not syntax:
                if "{%" in self.text_template or "{{" in self.text_template:
                    logger.info("Detected Jinja2 syntax in text template")
                    syntax = TemplateSyntaxEnum.jinja2
            if not syntax:
                syntax = TemplateSyntaxEnum.fstring
            if syntax == TemplateSyntaxEnum.jinja2:
                from jinja2 import Template

                template = Template(self.text_template)
                return template.render(**obj)
            elif syntax == TemplateSyntaxEnum.fstring:
                return self.text_template.format(**obj)
            else:
                raise NotImplementedError(f"Cannot handle template syntax: {syntax}")
        return str(obj)

    def search(
        self,
        query: str,
        vectors: List[Tuple[str, INDEX_ITEM]],
        limit: Optional[int] = None,
        mmr_relevance_factor: Optional[float] = None,
    ) -> List[Tuple[float, Any]]:
        """
        Use the indexer to search against a database of vectors.

        Note: this is a low-level method, typically you would use the :ref:`search` method on a :ref:`Collection`.

        :param query: The query string to search for
        :param vectors: A list of indexed items, where each item is a tuple of (id, vector)
        :param limit: The maximum number of results to return (optional)
        :return: A list of item IDs or objects that match the query
        """

        # Convert the query string to a vector
        query_vector = self.text_to_vector(query, cache=False)

        if mmr_relevance_factor is not None:
            vlist = [v for _, v in vectors]
            idlist = [id for id, _ in vectors]
            sorted_indices = mmr_diversified_search(
                query_vector, vlist, relevance_factor=mmr_relevance_factor, top_n=limit
            )
            results = []
            # TODO: this is inefficient when limit is high
            for i in range(limit):
                if i >= len(sorted_indices):
                    break
                pos = sorted_indices[i]
                score = pairwise_cosine_similarity(query_vector, vlist[pos])
                results.append((score, idlist[pos]))
            return results

        distances = []

        # Iterate over each indexed item
        for item_id, item_vector in vectors:
            # Calculate the Euclidean distance between the query vector and the item vector
            # distance = 1-np.linalg.norm(query_vector - item_vector)
            distance = pairwise_cosine_similarity(query_vector, item_vector)
            distances.append((distance, item_id))

        # Sort the distances in ascending order
        distances.sort(key=lambda x: -x[0])

        # Limit the number of results if specified
        if limit is not None:
            distances = distances[:limit]

        return distances



"""
Tests for the clustering functionality in heatmap visualization.
"""

import os
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from linkml_store.plotting.heatmap import create_heatmap


@pytest.fixture
def cluster_test_data():
    """Create a sample DataFrame designed for clustering tests."""
    # Create a dataset with clear patterns for clustering
    np.random.seed(42)  # For reproducibility
    
    # Create 20 rows x 15 columns with patterns
    n_rows = 20
    n_cols = 15
    
    # Base data - random
    data = np.random.rand(n_rows, n_cols)
    
    # Create patterns in the data
    # First 5 rows have similar pattern
    data[0:5, :] = data[0:5, :] * 0.5 + 0.8
    # Next 5 rows have another pattern
    data[5:10, :] = data[5:10, :] * 0.3 + 0.2
    # Next 5 rows have a third pattern
    data[10:15, :] = data[10:15, :] * 0.8 + 0.1
    # Last 5 rows have a fourth pattern
    data[15:20, :] = data[15:20, :] * 0.4 + 0.5
    
    # Similarly, create patterns in the columns
    data[:, 0:5] = data[:, 0:5] * 1.2
    data[:, 5:10] = data[:, 5:10] * 0.8
    data[:, 10:15] = data[:, 10:15] * 1.5
    
    # Convert to DataFrame
    row_names = [f"row_{i}" for i in range(n_rows)]
    col_names = [f"col_{i}" for i in range(n_cols)]
    df = pd.DataFrame(data, columns=col_names)
    
    # Add row and column identifiers
    df["row_id"] = row_names
    
    return df


def test_clustering_options(cluster_test_data):
    """Test that clustering options work as expected."""
    with tempfile.TemporaryDirectory() as temp_dir:
        # Test with no clustering (default)
        fig1, ax1 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",  # Any column name will work for this test
            y_column="row_id",
            value_column="col_1",  # Use another column as values
            cluster=False
        )
        
        # Test with clustering both axes
        fig2, ax2 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",
            y_column="row_id",
            value_column="col_1",
            cluster="both",
            cluster_method="complete",
            cluster_metric="euclidean"
        )
        
        # Test with clustering only x-axis
        fig3, ax3 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",
            y_column="row_id",
            value_column="col_1",
            cluster="x"
        )
        
        # Test with clustering only y-axis
        fig4, ax4 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",
            y_column="row_id",
            value_column="col_1",
            cluster="y"
        )
        
        # Test different linkage methods
        fig5, ax5 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",
            y_column="row_id",
            value_column="col_1",
            cluster="both",
            cluster_method="single"
        )
        
        # Test different distance metrics
        fig6, ax6 = create_heatmap(
            data=cluster_test_data,
            x_column="col_0",
            y_column="row_id",
            value_column="col_1",
            cluster="both",
            cluster_metric="correlation"
        )
        
        # Verify that all axes were created
        assert ax1 is not None
        assert ax2 is not None
        assert ax3 is not None
        assert ax4 is not None
        assert ax5 is not None
        assert ax6 is not None
        
        # Close all figures to avoid memory leaks
        import matplotlib.pyplot as plt
        plt.close('all')


"""Dremio REST API collection implementation.

This module provides the Collection implementation for Dremio REST API,
supporting query operations via the REST API v3.
"""

import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query, QueryResult

logger = logging.getLogger(__name__)


class DremioRestCollection(Collection):
    """Collection implementation for Dremio data lakehouse via REST API.

    This collection connects to Dremio tables via the REST API v3
    and provides query capabilities. Write operations may be limited
    depending on the underlying data source configuration in Dremio.
    """

    _table_exists_checked: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_table_path(self) -> str:
        """Get the full qualified table path.

        Returns:
            Full table path for SQL queries.
        """
        return self.parent._get_table_path(self.alias)

    def _build_select_sql(
        self,
        select_cols: Optional[List[str]] = None,
        where_clause: Optional[Union[str, Dict[str, Any]]] = None,
        sort_by: Optional[List[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> str:
        """Build a SELECT SQL statement.

        Args:
            select_cols: Columns to select (None for all).
            where_clause: WHERE conditions.
            sort_by: ORDER BY columns.
            limit: Maximum rows to return.
            offset: Number of rows to skip.

        Returns:
            SQL SELECT statement.
        """
        table_path = self._get_table_path()

        # Build SELECT clause
        if select_cols:
            cols = ", ".join(f'"{c}"' for c in select_cols)
        else:
            cols = "*"

        sql = f"SELECT {cols} FROM {table_path}"

        # Build WHERE clause
        if where_clause:
            conditions = self._build_where_conditions(where_clause)
            if conditions:
                sql += f" WHERE {conditions}"

        # Build ORDER BY clause
        if sort_by:
            order_cols = ", ".join(f'"{c}"' for c in sort_by)
            sql += f" ORDER BY {order_cols}"

        # Build LIMIT/OFFSET
        if limit is not None and limit >= 0:
            sql += f" LIMIT {limit}"
        if offset is not None and offset > 0:
            sql += f" OFFSET {offset}"

        return sql

    def _build_where_conditions(self, where_clause: Union[str, Dict[str, Any]]) -> str:
        """Build WHERE clause conditions from a dict or string.

        Args:
            where_clause: WHERE conditions as dict or string.

        Returns:
            SQL WHERE clause (without WHERE keyword).
        """
        if isinstance(where_clause, str):
            return where_clause

        if not isinstance(where_clause, dict):
            return ""

        conditions = []
        for key, value in where_clause.items():
            condition = self._build_single_condition(key, value)
            if condition:
                conditions.append(condition)

        return " AND ".join(conditions)

    def _build_single_condition(self, key: str, value: Any) -> str:
        """Build a single WHERE condition.

        Supports MongoDB-style operators like $gt, $gte, $lt, $lte, $in, $ne.

        Args:
            key: Column name.
            value: Value or operator dict.

        Returns:
            SQL condition string.
        """
        col = f'"{key}"'

        if value is None:
            return f"{col} IS NULL"

        if isinstance(value, dict):
            # Handle operators
            sub_conditions = []
            for op, val in value.items():
                if op == "$gt":
                    sub_conditions.append(f"{col} > {self._sql_value(val)}")
                elif op == "$gte":
                    sub_conditions.append(f"{col} >= {self._sql_value(val)}")
                elif op == "$lt":
                    sub_conditions.append(f"{col} < {self._sql_value(val)}")
                elif op == "$lte":
                    sub_conditions.append(f"{col} <= {self._sql_value(val)}")
                elif op == "$ne":
                    if val is None:
                        sub_conditions.append(f"{col} IS NOT NULL")
                    else:
                        sub_conditions.append(f"{col} != {self._sql_value(val)}")
                elif op == "$in":
                    if isinstance(val, (list, tuple)):
                        vals = ", ".join(self._sql_value(v) for v in val)
                        sub_conditions.append(f"{col} IN ({vals})")
                elif op == "$nin":
                    if isinstance(val, (list, tuple)):
                        vals = ", ".join(self._sql_value(v) for v in val)
                        sub_conditions.append(f"{col} NOT IN ({vals})")
                elif op == "$like":
                    sub_conditions.append(f"{col} LIKE {self._sql_value(val)}")
                elif op == "$ilike":
                    sub_conditions.append(f"LOWER({col}) LIKE LOWER({self._sql_value(val)})")
                elif op == "$regex":
                    # Dremio uses REGEXP_LIKE
                    sub_conditions.append(f"REGEXP_LIKE({col}, {self._sql_value(val)})")
                else:
                    logger.warning(f"Unknown operator: {op}")

            return " AND ".join(sub_conditions) if sub_conditions else ""
        else:
            return f"{col} = {self._sql_value(value)}"

    def _sql_value(self, value: Any) -> str:
        """Convert a Python value to SQL literal.

        Args:
            value: Python value.

        Returns:
            SQL literal string.
        """
        if value is None:
            return "NULL"
        elif isinstance(value, bool):
            return "TRUE" if value else "FALSE"
        elif isinstance(value, (int, float)):
            return str(value)
        elif isinstance(value, str):
            # Escape single quotes
            escaped = value.replace("'", "''")
            return f"'{escaped}'"
        elif isinstance(value, (list, dict)):
            # Convert to JSON string
            escaped = json.dumps(value).replace("'", "''")
            return f"'{escaped}'"
        else:
            escaped = str(value).replace("'", "''")
            return f"'{escaped}'"

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        """Insert objects into the collection.

        Note: Write operations in Dremio depend on the underlying data source.
        Some sources (like Iceberg, Delta Lake) support writes, while others
        (like file-based sources) may not.

        Args:
            objs: Object(s) to insert.
            **kwargs: Additional arguments.
        """
        if not isinstance(objs, list):
            objs = [objs]

        if not objs:
            return

        logger.debug(f"Inserting {len(objs)} objects into {self.alias}")

        cd = self.class_definition()
        if not cd:
            logger.debug(f"No class definition for {self.alias}; inducing from objects")
            cd = self.induce_class_definition_from_objects(objs)

        table_path = self._get_table_path()

        if cd and cd.attributes:
            columns = list(cd.attributes.keys())
        else:
            columns = list(objs[0].keys())

        col_list = ", ".join(f'"{c}"' for c in columns)

        batch_size = 100
        for i in range(0, len(objs), batch_size):
            batch = objs[i : i + batch_size]

            values_list = []
            for obj in batch:
                values = []
                for col in columns:
                    val = obj.get(col)
                    values.append(self._sql_value(val))
                values_list.append(f"({', '.join(values)})")

            values_sql = ", ".join(values_list)
            sql = f"INSERT INTO {table_path} ({col_list}) VALUES {values_sql}"

            self.parent._execute_update(sql)

        self._post_insert_hook(objs)

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> Optional[int]:
        """Delete specific objects from the collection.

        Args:
            objs: Object(s) to delete.
            **kwargs: Additional arguments.

        Returns:
            Number of deleted rows, or None if unknown.
        """
        if not isinstance(objs, list):
            objs = [objs]

        if not objs:
            return 0

        table_path = self._get_table_path()
        total_deleted = 0

        for obj in objs:
            conditions = []
            for key, value in obj.items():
                if key.startswith("_"):
                    continue
                condition = self._build_single_condition(key, value)
                if condition:
                    conditions.append(condition)

            if not conditions:
                continue

            sql = f"DELETE FROM {table_path} WHERE {' AND '.join(conditions)}"
            result = self.parent._execute_update(sql)
            if result > 0:
                total_deleted += result

        self._post_delete_hook()
        return total_deleted if total_deleted > 0 else None

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> Optional[int]:
        """Delete objects matching a condition.

        Args:
            where: WHERE conditions (empty dict means delete all).
            missing_ok: If True, don't raise error if no rows deleted.
            **kwargs: Additional arguments.

        Returns:
            Number of deleted rows, or None if unknown.
        """
        if where is None:
            where = {}

        table_path = self._get_table_path()

        if where:
            conditions = self._build_where_conditions(where)
            sql = f"DELETE FROM {table_path} WHERE {conditions}"
        else:
            sql = f"DELETE FROM {table_path}"

        result = self.parent._execute_update(sql)
        if result == 0 and not missing_ok:
            raise ValueError(f"No rows found for {where}")
        self._post_delete_hook()
        return result if result >= 0 else None

    def update(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        """Update objects in the collection.

        Note: Requires a primary key field to identify rows.

        Args:
            objs: Object(s) to update.
            **kwargs: Additional arguments.
        """
        if not isinstance(objs, list):
            objs = [objs]

        if not objs:
            return

        table_path = self._get_table_path()
        pk = self.identifier_attribute_name

        if not pk:
            raise ValueError("Cannot update without an identifier attribute")

        for obj in objs:
            if pk not in obj:
                raise ValueError(f"Object missing primary key field: {pk}")

            pk_value = obj[pk]

            set_parts = []
            for key, value in obj.items():
                if key == pk or key.startswith("_"):
                    continue
                set_parts.append(f'"{key}" = {self._sql_value(value)}')

            if not set_parts:
                continue

            set_clause = ", ".join(set_parts)
            sql = f'UPDATE {table_path} SET {set_clause} WHERE "{pk}" = {self._sql_value(pk_value)}'
            self.parent._execute_update(sql)

    def query(self, query: Query, **kwargs) -> QueryResult:
        """Execute a query against the collection.

        Args:
            query: Query specification.
            **kwargs: Additional arguments.

        Returns:
            QueryResult with matching rows.
        """
        self._pre_query_hook(query)

        limit = query.limit
        if limit == -1:
            limit = None

        sql = self._build_select_sql(
            select_cols=query.select_cols,
            where_clause=query.where_clause,
            sort_by=query.sort_by,
            limit=limit,
            offset=query.offset,
        )

        df = self.parent._execute_query(sql)

        # Convert DataFrame to list of dicts
        row_list = df.to_dict("records") if not df.empty else []

        # Get total count for pagination
        if query.offset or (limit is not None and len(row_list) == limit):
            count_sql = self._build_count_sql(query.where_clause)
            try:
                count_df = self.parent._execute_query(count_sql)
                total_rows = int(count_df.iloc[0, 0]) if not count_df.empty else len(row_list)
            except Exception:
                total_rows = len(row_list)
        else:
            total_rows = len(row_list)

        qr = QueryResult(query=query, num_rows=total_rows, rows=row_list, offset=query.offset or 0)

        if query.include_facet_counts and query.facet_slots:
            qr.facet_counts = self.query_facets(where=query.where_clause, facet_columns=query.facet_slots)

        return qr

    def _build_count_sql(self, where_clause: Optional[Union[str, Dict[str, Any]]] = None) -> str:
        """Build a COUNT SQL statement.

        Args:
            where_clause: WHERE conditions.

        Returns:
            SQL COUNT statement.
        """
        table_path = self._get_table_path()
        sql = f"SELECT COUNT(*) FROM {table_path}"

        if where_clause:
            conditions = self._build_where_conditions(where_clause)
            if conditions:
                sql += f" WHERE {conditions}"

        return sql

    def query_facets(
        self,
        where: Optional[Dict] = None,
        facet_columns: Optional[List[str]] = None,
        facet_limit: int = DEFAULT_FACET_LIMIT,
        **kwargs,
    ) -> Dict[Union[str, Tuple[str, ...]], List[Tuple[Any, int]]]:
        """Get facet counts for columns.

        Args:
            where: Filter conditions.
            facet_columns: Columns to get facets for.
            facet_limit: Maximum facet values per column.
            **kwargs: Additional arguments.

        Returns:
            Dictionary mapping column names to list of (value, count) tuples.
        """
        if facet_limit is None:
            facet_limit = DEFAULT_FACET_LIMIT

        results = {}
        cd = self.class_definition()
        table_path = self._get_table_path()

        if not facet_columns:
            if cd and cd.attributes:
                facet_columns = list(cd.attributes.keys())
            else:
                return results

        for col in facet_columns:
            if isinstance(col, tuple):
                col_list = ", ".join(f'"{c}"' for c in col)
                col_name = col
            else:
                col_list = f'"{col}"'
                col_name = col

            sql = f"SELECT {col_list}, COUNT(*) as cnt FROM {table_path}"

            if where:
                conditions = self._build_where_conditions(where)
                if conditions:
                    sql += f" WHERE {conditions}"

            sql += f" GROUP BY {col_list} ORDER BY cnt DESC"

            if facet_limit > 0:
                sql += f" LIMIT {facet_limit}"

            try:
                df = self.parent._execute_query(sql)

                facets = []
                for _, row in df.iterrows():
                    if isinstance(col, tuple):
                        value = tuple(row[c] for c in col)
                    else:
                        value = row[col]
                    count = int(row["cnt"])
                    facets.append((value, count))

                results[col_name] = facets

            except Exception as e:
                logger.warning(f"Facet query failed for {col}: {e}")
                results[col_name] = []

        return results

    def _check_if_initialized(self) -> bool:
        """Check if the collection's table exists.

        Returns:
            True if table exists.
        """
        if self._table_exists_checked:
            return True

        result = self.parent._table_exists(self.alias)
        if result:
            self._table_exists_checked = True
        return result



import csv
import json
import os

import pytest
import yaml

from tests import INPUT_DIR, OUTPUT_DIR


def pytest_configure(config):
    """Register custom markers."""
    config.addinivalue_line(
        "markers", "core: tests that work with minimal (no extras) install"
    )


def pytest_collection_modifyitems(config, items):
    """Auto-skip tests requiring optional dependencies when not installed."""
    # Check which optional deps are available
    optional_deps = {
        "mongodb": "pymongo",
        "neo4j": "neo4j",
        "ibis": "ibis",
        "llm": "llm",
        "sklearn": "sklearn",
        "matplotlib": "matplotlib",
    }

    missing_deps = {}
    for name, module in optional_deps.items():
        try:
            __import__(module)
        except ImportError:
            missing_deps[name] = pytest.mark.skip(
                reason=f"requires {module} (install with: pip install linkml-store[{name}])"
            )

# Ensure output directory exists for tests that write temp files
os.makedirs(OUTPUT_DIR, exist_ok=True)

TEST_DATA = [
    {"id": 1, "name": "John", "age": 30},
    {"id": 2, "name": "Alice", "age": 25},
    {"id": 3, "name": "Bob", "age": 35},
]
TEST_DIR = INPUT_DIR / "test_data"
JSON_FILE = os.path.join(TEST_DIR, "data.json")
YAML_FILE = os.path.join(TEST_DIR, "data.yaml")
TSV_FILE = os.path.join(TEST_DIR, "data.tsv")
CSV_FILE = os.path.join(TEST_DIR, "data.csv")

os.makedirs(TEST_DIR, exist_ok=True)
with open(JSON_FILE, "w") as f:
    json.dump(TEST_DATA, f)
with open(YAML_FILE, "w") as f:
    yaml.safe_dump(TEST_DATA, f, sort_keys=False)
with open(TSV_FILE, "w") as f:
    writer = csv.DictWriter(f, fieldnames=TEST_DATA[0].keys(), delimiter="\t")
    writer.writeheader()
    writer.writerows(TEST_DATA)
with open(CSV_FILE, "w") as f:
    writer = csv.DictWriter(f, fieldnames=TEST_DATA[0].keys())
    writer.writeheader()
    writer.writerows(TEST_DATA)



import json
import matplotlib.pyplot as plt
import numpy as np
from collections import Counter

def create_faceted_horizontal_barchart(data, output_path, 
                                     figsize_per_facet=(8, 4)):
    """
    Create horizontal bar charts for each facet in the data.
    
    Args:
        data: Dictionary where keys are facet names and values are lists of items
        output_path: Path to save the PNG file
        figsize_per_facet: (width, height) for each subplot
    """
    
    # Calculate figure size based on number of facets
    new_data = {}
    for k, fm in data.items():
        new_fm = {str(k)[0:50]: v for k, v in list(fm.items())[0:10] if isinstance(k, (str, int, float))}
        new_data[k] = new_fm
    data = new_data
    n_facets = len(data)
    fig_width = figsize_per_facet[0]
    fig_height = figsize_per_facet[1] * n_facets
    
    # Create subplots
    fig, axes = plt.subplots(n_facets, 1, figsize=(fig_width, fig_height))
    
    # Handle case where there's only one facet
    if n_facets == 1:
        axes = [axes]
    
    # Process each facet
    for i, (facet_name, items) in enumerate(data.items()):
        ax = axes[i]
        
        # Count occurrences of each item
        counts = Counter(items)
        
        # Sort by count (descending) for better visualization
        sorted_items = sorted(counts.items(), key=lambda x: x[1], reverse=True)
        
        # Separate labels and values
        labels = [item[0] for item in sorted_items]
        values = [item[1] for item in sorted_items]
        
        # Create horizontal bar chart
        y_pos = np.arange(len(labels))
        bars = ax.barh(y_pos, values, alpha=0.7)
        
        # Customize the subplot
        ax.set_yticks(y_pos)
        ax.set_yticklabels(labels)
        ax.set_xlabel('Count')
        ax.set_title(f'{facet_name.replace("_", " ").title()}')
        ax.grid(axis='x', alpha=0.3)
        
        # Add value labels on bars
        for j, (bar, value) in enumerate(zip(bars, values)):
            ax.text(bar.get_width() + 0.01 * max(values), bar.get_y() + bar.get_height()/2, 
                   str(value), ha='left', va='center', fontsize=9)
        
        # Invert y-axis so highest counts are at top
        ax.invert_yaxis()
    
    # Adjust layout to prevent overlap
    plt.tight_layout()
    
    # Save the plot
    plt.savefig(output_path, dpi=150, bbox_inches='tight')
    # print(f"Faceted bar chart saved to: {output_path}")
    



import json
import logging
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
import sqlalchemy
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition
from linkml_runtime.utils.schema_builder import SchemaBuilder
from sqlalchemy import NullPool, text

from linkml_store.api import Database
from linkml_store.api.queries import Query, QueryResult
from linkml_store.api.stores.duckdb.duckdb_collection import DuckDBCollection
from linkml_store.utils.format_utils import Format
from linkml_store.utils.sql_utils import introspect_schema, query_to_sql

TYPE_MAP = {
    "VARCHAR": "string",
    "BIGINT": "integer",
    "BOOLEAN": "boolean",
    "DATE": "date",
    "DOUBLE": "float",
    "INTEGER": "integer",
    "JSON": "Any",
}

MEMORY_HANDLE = "duckdb:///:memory:"


logger = logging.getLogger(__name__)


class DuckDBDatabase(Database):
    """
    An adapter for DuckDB databases.

    Note that this adapter does not make use of a LinkML relational model transformation and
    SQL Alchemy ORM layer. Instead, it attempts to map each collection (which is of type
    some LinkML class) to a *single* DuckDB table. New tables are not created for nested references,
    and linking tables are not created for many-to-many relationships.

    Instead the native DuckDB ARRAY type is used to store multivalued attributes, and DuckDB JSON
    types are used for nested inlined objects.
    """

    # _connection: DuckDBPyConnection = None
    _engine: sqlalchemy.Engine = None
    collection_class = DuckDBCollection

    def __init__(self, handle: Optional[str] = None, recreate_if_exists: bool = False, **kwargs):
        if handle is None:
            handle = MEMORY_HANDLE
        if recreate_if_exists:
            path = Path(handle.replace("duckdb:///", ""))
            if path.exists():
                path.unlink()
        super().__init__(handle=handle, **kwargs)

    @property
    def engine(self) -> sqlalchemy.Engine:
        if not self._engine:
            handle = self.handle
            if not handle.startswith("duckdb://") and not handle.startswith(":") and "://" not in handle:
                handle = f"duckdb:///{handle}"
            if ":memory:" not in handle:
                # TODO: investigate this; duckdb appears to be prematurely caching
                self._engine = sqlalchemy.create_engine(handle, poolclass=NullPool)
            else:
                self._engine = sqlalchemy.create_engine(handle)
        return self._engine

    @property
    def _is_sqlite(self) -> bool:
        return self.handle and self.handle.startswith("sqlite:")

    def commit(self, **kwargs):
        with self.engine.connect() as conn:
            conn.commit()

    def close(self, **kwargs):
        self.engine.dispose()

    def drop(self, missing_ok=True, **kwargs):
        self.close()
        if self.handle == MEMORY_HANDLE:
            return
        path = Path(self.handle.replace("duckdb:///", ""))
        if path.exists():
            path.unlink()
        else:
            if not missing_ok:
                raise FileNotFoundError(f"Database file not found: {path}")

    def _table_exists(self, table: str) -> bool:
        if self._is_sqlite:
            if table == "sqlite_master":
                return True
            meta_query = Query(
                from_table="sqlite_master",
                where_clause={
                    # "type": "table",
                    "name": table,
                },
            )
        else:
            if table.startswith("information_schema"):
                return True
            meta_query = Query(
                from_table="information_schema.tables",
                where_clause={
                    "table_type": "BASE TABLE",
                    "table_name": table,
                },
            )

        qr = self.query(meta_query)
        if qr.num_rows == 0:
            logger.debug(f"Table {self.alias} not created yet")
            return False
        return True

    def _json_encoded_cols(self, table_name: str) -> Optional[List[str]]:
        json_encoded_cols = []
        if table_name:
            if table_name.startswith("information_schema") or table_name.startswith("sqlite"):
                return []
            sv = self.schema_view
            if sv:
                cd = None
                for c in self._collections.values():
                    if c.alias == table_name or c.target_class_name == table_name:
                        cd = c.class_definition()
                        break
                if cd:
                    for att in sv.class_induced_slots(cd.name):
                        if att.inlined or att.inlined_as_list:
                            json_encoded_cols.append(att.name)
        return json_encoded_cols

    def query(self, query: Query, **kwargs) -> QueryResult:
        if not self._table_exists(query.from_table):
            return QueryResult(query=query, num_rows=0, rows=[])
        json_encoded_cols = self._json_encoded_cols(query.from_table)

        with self.engine.connect() as conn:
            count_query_str = text(query_to_sql(query, count=True))
            logger.debug(f"count_query_str: {count_query_str}")
            num_rows = list(conn.execute(count_query_str))[0][0]
            logger.debug(f"num_rows: {num_rows}")
            query_str = query_to_sql(query, **kwargs)  # include offset, limit
            logger.debug(f"query_str: {query_str}")
            rows = list(conn.execute(text(query_str)).mappings())
            qr = QueryResult(query=query, num_rows=num_rows, rows=rows)
            if json_encoded_cols:
                for row in qr.rows:
                    for col in json_encoded_cols:
                        if row[col]:
                            if isinstance(row[col], list):
                                for i in range(len(row[col])):
                                    try:
                                        parsed_val = json.loads(row[col][i])
                                    except json.JSONDecodeError as e:
                                        logger.error(f"Failed to parse col {col}[{i}] == {row[col][i]}")
                                        raise e
                                    row[col][i] = parsed_val
                            elif isinstance(row[col], dict):
                                pass
                            else:
                                row[col] = json.loads(row[col])
            qr.set_rows(pd.DataFrame(rows))
            facet_columns = query.facet_slots
            if query.include_facet_counts and not facet_columns:
                raise ValueError("Facet counts requested but no facet columns specified")
            if facet_columns:
                raise NotImplementedError
            return qr

    @property
    def supports_sql(self) -> bool:
        """Return True - DuckDB supports raw SQL queries."""
        return True

    def execute_sql(self, sql: str, **kwargs) -> QueryResult:
        """
        Execute a raw SQL query against the DuckDB database.

        :param sql: SQL query string
        :param kwargs: Additional arguments
        :return: QueryResult containing the results
        """
        with self.engine.connect() as conn:
            result = conn.execute(text(sql))
            rows = [dict(row._mapping) for row in result]
            return QueryResult(num_rows=len(rows), rows=rows)

    def init_collections(self):
        # TODO: unify schema introspection
        if not self.schema_view:
            schema = introspect_schema(self.engine)
        else:
            schema = self.schema_view.schema
        table_names = schema.classes.keys()
        if self._collections is None:
            self._collections = {}
        for table_name in table_names:
            if table_name not in self._collections:
                collection = DuckDBCollection(name=table_name, parent=self)
                self._collections[table_name] = collection

    def induce_schema_view(self) -> SchemaView:
        # TODO: unify schema introspection
        # TODO: handle case where schema is provided in advance
        logger.info(f"Inducing schema view for {self.metadata.handle} // {self}")
        sb = SchemaBuilder()
        schema = sb.schema
        logger.info(f"Checking if {self.metadata.handle} is sqlite: {self._is_sqlite}")
        if self._is_sqlite:
            return SchemaView(schema)
        query = Query(from_table="information_schema.tables", where_clause={"table_type": "BASE TABLE"})
        qr = self.query(query)
        logger.info(f"Found {qr.num_rows} information_schema.tables // {qr.rows}")
        if qr.num_rows:
            table_names = [row["table_name"] for row in qr.rows]
            for tbl in table_names:
                sb.add_class(tbl)
        query = Query(from_table="information_schema.columns", sort_by=["ordinal_position"])
        for row in self.query(query, limit=-1).rows:
            tbl_name = row["table_name"]
            if tbl_name not in sb.schema.classes:
                continue
            dt = row["data_type"]
            if dt.endswith("[]"):
                dt = dt[0:-2]
                multivalued = True
            else:
                multivalued = False
            rng = TYPE_MAP.get(dt, "string")
            sd = SlotDefinition(
                row["column_name"], required=row["is_nullable"] == "NO", multivalued=multivalued, range=rng
            )
            if dt == "JSON":
                sd.inlined_as_list = True
            sb.schema.classes[tbl_name].attributes[sd.name] = sd
            logger.info(f"Introspected slot: {tbl_name}.{sd.name}: {sd.range} FROM {dt}")
        sb.add_defaults()
        for cls_name in schema.classes:
            if cls_name in self.metadata.collections:
                collection_metadata = self.metadata.collections[cls_name]
                if collection_metadata.attributes:
                    del schema.classes[cls_name]
                    cls = ClassDefinition(name=collection_metadata.type, attributes=collection_metadata.attributes)
                    schema.classes[cls.name] = cls
        return SchemaView(schema)

    def export_database(self, location: str, target_format: Optional[Union[str, Format]] = None, **kwargs):
        if target_format == "duckdb" or target_format == Format.SQLDUMP_DUCKDB:
            path = Path(location)
            if path.exists():
                if path.is_file():
                    path.unlink()
            with self.engine.connect() as conn:
                sql = text(f"EXPORT DATABASE '{location}'")
                conn.execute(sql)
        else:
            super().export_database(location, target_format=target_format, **kwargs)

    def import_database(self, location: str, source_format: Optional[str] = None, **kwargs):
        """
        Import a database from a file or location.

        :param location: location of the file
        :param source_format: source format
        :param kwargs: additional arguments
        """
        if source_format == Format.SQLDUMP_DUCKDB.value or source_format == Format.SQLDUMP_DUCKDB:
            with self.engine.connect() as conn:
                sql = text(f"IMPORT DATABASE '{location}'")
                conn.execute(sql)
                conn.commit()
        else:
            super().import_database(location, source_format=source_format, **kwargs)



import logging
from typing import Any, List, Optional, Tuple

from pydantic import BaseModel, ConfigDict, Field

from linkml_store.api.collection import OBJECT
from linkml_store.utils.format_utils import Format, load_objects

logger = logging.getLogger(__name__)


class LLMConfig(BaseModel, extra="forbid"):
    """
    Configuration for the LLM indexer.
    """

    model_config = ConfigDict(protected_namespaces=())

    model_name: str = "gpt-4o-mini"
    token_limit: Optional[int] = None
    number_of_few_shot_examples: Optional[int] = None
    role: str = "Domain Expert"
    cached_embeddings_database: Optional[str] = None
    cached_embeddings_collection: Optional[str] = None
    text_template: Optional[str] = None
    text_template_syntax: Optional[str] = None


class InferenceConfig(BaseModel, extra="forbid"):
    """
    Configuration for inference engines.
    """

    target_attributes: Optional[List[str]] = None
    feature_attributes: Optional[List[str]] = None
    train_test_split: Optional[Tuple[float, float]] = None
    llm_config: Optional[LLMConfig] = None
    random_seed: Optional[int] = None
    validate_results: Optional[bool] = None

    @classmethod
    def from_file(cls, file_path: str, format: Optional[Format] = None) -> "InferenceConfig":
        """
        Load an inference config from a file.

        :param file_path: Path to the file.
        :param format: Format of the file (YAML is recommended).
        :return: InferenceConfig
        """
        if format and format.is_xsv():
            logger.warning("XSV format is not recommended for inference config files")
        objs = load_objects(file_path, format=format)
        if len(objs) != 1:
            raise ValueError(f"Expected 1 object, got {len(objs)}")
        return cls(**objs[0])


class Inference(BaseModel, extra="forbid"):
    """
    Result of an inference derivation.
    """

    query: Optional[OBJECT] = Field(default=None, description="The query object.")
    predicted_object: OBJECT = Field(..., description="The predicted object.")
    confidence: Optional[float] = Field(default=None, description="The confidence of the prediction.", le=1.0, ge=0.0)
    explanation: Optional[Any] = Field(default=None, description="Explanation of the prediction.")



import pytest

from linkml_store.utils.object_utils import object_path_update


@pytest.mark.parametrize(
    "input,path,value,expected",
    [
        ({}, "foo", 1, {"foo": 1}),
        ({"foo": 2}, "foo", 1, {"foo": 1}),
        ({"foo": None}, "foo", 1, {"foo": 1}),
        ({"foo": {"bar": 1}}, "foo", 1, {"foo": 1}),
        ({}, "foo.bar", 1, {"foo": {"bar": 1}}),
        ({"foo": {}}, "foo.bar", 1, {"foo": {"bar": 1}}),
        ({"foo": None}, "foo.bar", 1, {"foo": {"bar": 1}}),
        ({"foo": {"xyz": 2}}, "foo.bar", 1, {"foo": {"xyz": 2, "bar": 1}}),
        ({"foo": [2]}, "foo[0]", 1, {"foo": [1]}),
        ({}, "persons[0].foo.bar", 1, {"persons": [{"foo": {"bar": 1}}]}),
        ({}, "persons[1].foo.bar", 1, {"persons": [{}, {"foo": {"bar": 1}}]}),
        ({"persons": [{"foo": {"bar": 1}}]}, "persons[0].foo.bar", 1, {"persons": [{"foo": {"bar": 1}}]}),
        ({"persons": [{"foo": {"bar": 1}}]}, "persons[0].foo.baz", 2, {"persons": [{"foo": {"bar": 1, "baz": 2}}]}),
        (
            {"persons": [{"foo": {"bar": 1}}]},
            "persons[1].foo.baz",
            2,
            {"persons": [{"foo": {"bar": 1}}, {"foo": {"baz": 2}}]},
        ),
    ],
)
def test_update_nested_object(input, path, value, expected):
    result = object_path_update(input, path, value)
    assert result == expected



"""Type mappings between LinkML types and Dremio/Arrow types."""

import pyarrow as pa

# Mapping from LinkML types to PyArrow types
LINKML_TO_ARROW = {
    "string": pa.string(),
    "integer": pa.int64(),
    "float": pa.float64(),
    "boolean": pa.bool_(),
    "date": pa.date32(),
    "datetime": pa.timestamp("us"),
    "decimal": pa.decimal128(38, 10),
    "Any": pa.string(),  # Fallback to string for Any type
}

# Mapping from Arrow types to LinkML types
ARROW_TO_LINKML = {
    pa.string(): "string",
    pa.utf8(): "string",
    pa.large_string(): "string",
    pa.int8(): "integer",
    pa.int16(): "integer",
    pa.int32(): "integer",
    pa.int64(): "integer",
    pa.uint8(): "integer",
    pa.uint16(): "integer",
    pa.uint32(): "integer",
    pa.uint64(): "integer",
    pa.float16(): "float",
    pa.float32(): "float",
    pa.float64(): "float",
    pa.bool_(): "boolean",
    pa.date32(): "date",
    pa.date64(): "date",
}

# Mapping from Dremio SQL type names to LinkML types
DREMIO_SQL_TO_LINKML = {
    "VARCHAR": "string",
    "CHAR": "string",
    "BIGINT": "integer",
    "INTEGER": "integer",
    "INT": "integer",
    "SMALLINT": "integer",
    "TINYINT": "integer",
    "BOOLEAN": "boolean",
    "DOUBLE": "float",
    "FLOAT": "float",
    "DECIMAL": "float",
    "DATE": "date",
    "TIMESTAMP": "datetime",
    "TIME": "string",
    "BINARY": "string",
    "VARBINARY": "string",
    "LIST": "string",  # Complex types mapped to string
    "STRUCT": "string",
    "MAP": "string",
}


def get_arrow_type(linkml_type: str) -> pa.DataType:
    """Convert a LinkML type to a PyArrow type.

    Args:
        linkml_type: The LinkML type name.

    Returns:
        The corresponding PyArrow data type.
    """
    return LINKML_TO_ARROW.get(linkml_type, pa.string())


def get_linkml_type_from_arrow(arrow_type: pa.DataType) -> str:
    """Convert a PyArrow type to a LinkML type.

    Args:
        arrow_type: The PyArrow data type.

    Returns:
        The corresponding LinkML type name.
    """
    # Handle parameterized types by checking base type
    if pa.types.is_string(arrow_type) or pa.types.is_large_string(arrow_type):
        return "string"
    if pa.types.is_integer(arrow_type):
        return "integer"
    if pa.types.is_floating(arrow_type):
        return "float"
    if pa.types.is_boolean(arrow_type):
        return "boolean"
    if pa.types.is_date(arrow_type):
        return "date"
    if pa.types.is_timestamp(arrow_type):
        return "datetime"
    if pa.types.is_decimal(arrow_type):
        return "float"
    if pa.types.is_list(arrow_type) or pa.types.is_large_list(arrow_type):
        return "string"  # Complex types as string
    if pa.types.is_struct(arrow_type):
        return "string"
    if pa.types.is_map(arrow_type):
        return "string"

    return "string"  # Default fallback
